import enum
import logging
import asyncio
import os

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.MSSQL import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.core.logging.log_objects import Credential



class MSSQLSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.status = SessionStatus.START

class MSSQL(ResponderServer):
	def init(self):
		pass

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[TDSPacket.from_streamreader(self.creader, timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
				
			if isinstance(result[0], R3ConnectionClosed):
				return
				
			elif isinstance(result[0], Exception):
				raise result[0]
				
			else:
				packet = result[0]

			#print(str(packet))

			if self.session.status == SessionStatus.START:
				if packet.type != PacketType.PRELOGIN:
					raise Exception('Unexpected packet type! %s '% packet.type)
				self.session.status = SessionStatus.PREAUTH_SENT
				data = PRELOGIN()
				data.version = b'\x11\x02\x00\x00\x00\x00'
				data.encryption = Encryption.NOT_SUP
				data.instvalidity = ''
				data.thread_id = 0
				data.mars = False
				data.fedauthrequired = False
				data.traceid = os.urandom(32)

				rp = TDSPacket()
				rp.type = PacketType.TABULAR_RESULT
				rp.status = PacketStatus.EOM
				rp.spid = 0
				rp.packet_id = 1
				rp.window = 0
				rp.data = data
				await self.send_data(rp.to_bytes())
				continue

			elif self.session.status == SessionStatus.PREAUTH_SENT:
				if packet.type == PacketType.LOGIN7:
					cred = Credential(
							'plaintext',
							username = packet.data.username,
							password = packet.data.password,
							fullhash = '%s:%s' % (packet.data.username, packet.data.password)
						)
					await self.logger.credential(cred)
					return

				elif packet.type == PacketType.SSPI:
					self.session.status = SessionStatus.SSPI_AUTH
					#TODO
					return
					#raise Exception('Not implemented!')

				#elif 

			#TODO: implement SSPI and SSL auth (latter is encryption on flag set)
			#elif self.session.status == SessionStatus.SSPI_AUTH:

			else:
				raise Exception('Unexpected packet at this stage!')




import os
from abc import ABC, abstractmethod
import threading
import logging
import logging.config
import asyncio
import traceback
import sys
import importlib
import uuid
from pathlib import Path
import io

from responder3.core.logging.logger import *
from responder3.core.logging.log_objects import *
from responder3.core.commons import *
from responder3.core.manager.comms import *


class LogExtensionTaskEntry:
	def __init__(self, taskid):
		self.taskid = taskid
		self.created_at = datetime.datetime.utcnow()
		self.started_at = None
		self.extension_config = None
		self.extension_coro = None
		self.extension_handler = None
		self.extension_log_queue = asyncio.Queue()
		self.extension_command_queue = asyncio.Queue()


class LogProcessor:
	def __init__(self, log_settings, log_queue, loop = None, manager_log_queue = None):
		"""
		Extensible logging process. Does the logging via python's built-in logging module.
		:param logsettings: Dictionary describing the logging settings
		:type logsettings: dict
		:param logQ: Queue to read logging messages from
		:type logQ: multiprocessing.Queue
		"""
		self.loop = loop if loop is not None else asyncio.get_event_loop()
		self.log_settings = log_settings
		self.log_queue = log_queue
		self.manager_log_queue = manager_log_queue
		self.logger = None
		self.extensions_tasks = {}
		self.extension_task_id = 0
		self.result_history = {}
		self.proxy_file_handler = None
		self.name = 'LogProcessor'

	async def log(self, message, level=logging.INFO):
		"""
		Logging function used to send logs in this process only!
		:param message: The message to be logged
		:type message: str
		:param level: log level
		:type level: int
		:return: None
		"""
		await self.handle_log(LogEntry(level, self.name, message))

	async def start_extension(self, handler, extension_config):
		try:
			let = LogExtensionTaskEntry(self.extension_task_id)
			self.extension_task_id += 1
			let.extension_config = extension_config
			let.extension_handler = handler(
				self.log_queue,
				let.extension_log_queue,
				let.extension_command_queue,
				let.extension_config,
				self.loop
			)
			let.extension_coro = let.extension_handler.create_coro()
			self.extensions_tasks[let.taskid] = let
			self.loop.create_task(let.extension_coro)
		except Exception as e:
			await self.log_exception()

		return

	def create_dir_strucutre(self):
		if 'logdir' in self.log_settings:
			Path(self.log_settings['logdir']).mkdir(parents=True, exist_ok=True)
			Path(self.log_settings['logdir'], 'emails').mkdir(parents=True, exist_ok=True)
			Path(self.log_settings['logdir'], 'proxydata').mkdir(parents=True, exist_ok=True)
			Path(self.log_settings['logdir'], 'poisondata').mkdir(parents=True, exist_ok=True)
			Path(self.log_settings['logdir'], 'creds').mkdir(parents=True, exist_ok=True)

	async def get_handlers(self):
		for handler in self.log_settings['handlers']:
			if handler == 'TEST':
				handlerclass = TestExtension
				handlerclassname = 'TEST'
			else:
				handlerclassname = '%sHandler' % handler
				handlermodulename = 'responder3_log_%s' % handler.replace('-', '_').lower()
				#handlermodulename = '%s.%s' % (handlermodulename, handlerclassname)
				await self.log('Importing handler module: %s , %s' % (handlermodulename, handlerclassname), logging.DEBUG)
				handlerclass = getattr(importlib.import_module(handlermodulename, handlerclassname), handlerclassname)

			yield handlerclass, handler, handlermodulename

	async def setup(self):
		"""
		Parses the settings dict and populates the necessary variables
		:return: None
		"""
		logging.config.dictConfig(self.log_settings['log'])
		self.logger = logging.getLogger('Responder3')
		self.create_dir_strucutre()

		if 'handlers' in self.log_settings:
			async for handlerclass, handler, handlermodulename in self.get_handlers():
				for handler_config_name in self.log_settings['handlers'][handler]:
					await self.log('Starting log plugin %s with config name %s' % (handlermodulename, handler_config_name), logging.DEBUG)
					if handler_config_name not in self.log_settings:
						raise Exception('Failed to load log plugin %s! Reason: Conefig name %s is not in the config file!' % (handlermodulename, handler_config_name))
					await self.start_extension(handlerclass, self.log_settings[handler_config_name])

	async def run(self):
		try:
			await self.setup()
			await self.log('setup done', logging.DEBUG)
			while True:
				result = await self.log_queue.get()
				if not isinstance(result, R3CliLog): #not sending unserialized remotelog format to plugins
					for taskid in self.extensions_tasks:
						#await self.extensions_tasks[taskid].extension_log_queue.put(result)
						
						"""
						not using await for filling up the extension input queues, because if one extension malfuntions
						it stops the whole application by calling await on a full queue thet will never be empty
						"""
						try:
							self.extensions_tasks[taskid].extension_log_queue.put_nowait(result)
						except asyncio.QueueFull:
							pass
				if self.manager_log_queue:
					await self.manager_log_queue.put(result)
				if isinstance(result, Credential):
					await self.handle_credential(result)
				elif isinstance(result, LogEntry):
					await self.handle_log(result)
				elif isinstance(result, (Connection, ConnectionOpened, ConnectionClosed)):
					await self.handle_connection(result)
				elif isinstance(result, EmailEntry):
					await self.handle_email(result)
				elif isinstance(result, PoisonResult):
					await self.handle_poisonresult(result)
				elif isinstance(result, ProxyData):
					await self.handle_proxydata(result)
				elif isinstance(result, TrafficLog):
					await self.handle_traffic(result)
				elif isinstance(result, (R3CliLog, RemoteLog)):
					await self.handle_remote_log(result)
				else:
					raise Exception('Unknown object in queue! Got type: %s' % type(result))

		except Exception as e:
			await self.log_exception('Logger main task exception')

		finally:
			if self.proxy_file_handler is not None:
				self.proxy_file_handler.close()

	async def handle_log(self, log):
		"""
		Handles the messages of log type
		:param log: Log message object
		:type log: LogEntry
		:return: None
		"""
		self.logger.log(log.level, str(log))

	async def handle_connection(self, con):
		"""
		Handles the messages of log type
		:param con: Connection message object
		:type con: Connection
		:return: None
		"""
		if isinstance(con, Connection):
			self.logger.log(logging.INFO, str(con))
		
		elif isinstance(con, ConnectionOpened):
			self.logger.log(logging.INFO, str(con))
			
		elif isinstance(con, ConnectionClosed):
			self.logger.log(logging.INFO, str(con))

	async def handle_traffic(self,traffic):
		for line in traffic.get_loglines():
			self.logger.log(logging.INFO, line)

	async def handle_credential(self, result):
		"""
		Logs credential object arriving from logqueue
		:param result: Credential object to log
		:type result: Credential
		:return: None
		"""
		if 'logdir' in self.log_settings:
			filename = 'cred_%s_%s.json' % (datetime.datetime.utcnow().isoformat(), str(uuid.uuid4()))
			with open(str(Path(self.log_settings['logdir'], 'creds', filename).resolve()), 'wb') as f:
				f.write(result.to_json())

		# TODO: fingerprinting to avoid duplicate results!!!!
		"""
		if result.fingerprint not in self.result_history:
			await self.log(str(result.to_dict()), logging.INFO)
			self.result_history[result.fingerprint] = result
		else:
			await self.log('Duplicate result found! Filtered.')
		"""
		#line = '%s[%s] %s' % (result.module, 'CREDENTIAL', result.fullhash)
		self.logger.log(logging.INFO, str(result))
		#await self.log(str(result.to_dict()), logging.INFO)
			
	async def handle_email(self, email):
		"""
		Logs the email object arriving from logqueue
		:param email: Email object to log
		:type email: Email
		:return:
		"""
		if 'logdir' in self.log_settings:
			filename = 'email_%s_%s.eml' % (datetime.datetime.utcnow().isoformat(), str(uuid.uuid4()))
			with open(str(Path(self.log_settings['logdir'], 'emails', filename).resolve()), 'wb') as f:
				f.write(email.email.as_bytes())

		await self.log('You got mail!')

	async def handle_poisonresult(self, poisonresult):
		"""
		Logs the poisonresult object arriving from logqueue
		:param poisonresult:
		:type poisonresult: PoisonResult
		:return: None
		"""
		if 'logdir' in self.log_settings:
			filename = 'pr_%s_%s.json' % (datetime.datetime.utcnow().isoformat(), str(uuid.uuid4()))
			with open(str(Path(self.log_settings['logdir'], 'poisondata', filename).resolve()), 'wb') as f:
				f.write(poisonresult.to_json())
		await self.log(repr(poisonresult))

	async def handle_proxydata(self, proxydata):
		# TODO: currently it flushes everything on each line, this is not good (slow)
		# need to write a better scheduler for outout, timer maybe?
		"""
		Writes the incoming proxydata to a file
		:param proxydata: ProxyData
		:type proxydata: ProxyData
		:return: None
		"""
		if 'logdir' in self.log_settings and self.proxy_file_handler is None:
			filename = 'pr_%s_%s.json' % (datetime.datetime.utcnow().isoformat(), str(uuid.uuid4()))
			self.proxy_file_handler = open(str(Path(self.log_settings['logdir'], 'proxydata', filename).resolve()), 'wb')

		if self.proxy_file_handler is not None:
			try:
				self.proxy_file_handler.write(proxydata.to_json().encode() + b'\r\n')
				self.proxy_file_handler.flush()
				os.fsync(self.proxy_file_handler.fileno())
			except Exception as e:
				await self.log_exception('Error writing proxy data to file!')
				return
		await self.log(repr(proxydata), logging.DEBUG)
		
	async def handle_remote_log(self, remotelog):
		"""
		Deserializes remote log object and creates a local log object,
		then inserts the new object back to the logqueue
		"""
		try:
			if isinstance(remotelog, RemoteLog):
				if isinstance(remotelog.log_obj, LogEntry):
					self.logger.log(remotelog.log_obj.level, '[%s]%s' % ('REMOTE',str(remotelog)))
				elif isinstance(remotelog.log_obj, TrafficLog):
					for line in remotelog.log_obj.get_loglines():
						self.logger.log(logging.INFO, '[%s]%s' % ('REMOTE',line))

				else:
					self.logger.log(logging.INFO, '[%s]%s' % ('REMOTE',str(remotelog)))
			else:
				await self.log_queue.put(RemoteLog(remotelog))
		except Exception as e:
			await self.log_exception('handle_remote_log')
		
	# this function is a duplicate, clean it up!
	async def log_exception(self, message=None):
		"""
		Custom exception handler to log exceptions via the logging interface
		:param message: Extra message for the exception if any
		:type message: str
		:return: None
		"""
		sio = io.StringIO()
		ei = sys.exc_info()
		tb = ei[2]
		traceback.print_exception(ei[0], ei[1], tb, None, sio)
		msg = sio.getvalue()
		if msg[-1] == '\n':
			msg = msg[:-1]
		sio.close()
		if message is not None:
			msg = message + msg
		await self.log(msg, level=logging.ERROR)


class LoggerExtensionTask(ABC):
	def __init__(self, log_queue, result_queue, command_queue, config, loop):
		self.result_queue = result_queue
		self.log_queue = log_queue
		self.command_queue = command_queue
		self.loop = loop
		self.config = config
		self.modulename = '%s-%s' % ('LogExt', self.__class__.__name__)
		self.logger = Logger(self.modulename, logQ = self.log_queue)
		self.init()

	async def run(self):
		try:
			await self.setup()
			await self.logger.debug('Started!')
			await self.main()
			await self.logger.debug('Exiting!')
		except Exception:
			await self.logger.exception('Exception in main function!')

	
	async def create_coro(self):
		return await self.main()
			
	@abstractmethod
	def init(self):
		pass

	@abstractmethod
	async def main(self):
		pass

	@abstractmethod
	async def setup(self):
		pass


class TestExtension(LoggerExtensionTask):
	def init(self):
		self.output_queue = self.config['output_queue']

	async def setup(self):
		pass

	async def create_coro(self):
		return await self.main()

	async def main(self):
		try:
			while True:
				result = await self.result_queue.get()
				self.output_queue.put(result)
		except Exception as e:
			await self.logger.exception()



# https://msdn.microsoft.com/en-us/library/cc240469.aspx
import enum


class RDP_PROTOCOL(enum.IntFlag):
	PROTOCOL_RDP = 0x00
	PROTOCOL_SSL = 0x01
	PROTOCOL_HYBRID = 0x02
	PROTOCOL_RDSTLS = 0x04
	PROTOCOL_HYBRID_EX = 0x08

################################  Client X.224 Connection Request PDU ############################
class RDP_NEG_REQ_FAGS(enum.IntFlag):
	RESTRICTED_ADMIN_MODE_REQUIRED = 0x01
	REDIRECTED_AUTHENTICATION_MODE_REQUIRED = 0x02
	CORRELATION_INFO_PRESENT = 0x08

class RDP_NEG_REQ:
	def __init__(self):
		self.type = None
		self.flags = None
		self.length = None
		self.requestedProtocols = None

class RDP_NEG_CORRELATION_INFO:
	def __init__(self):
		self.type = None
		self.flags = None
		self.length = None
		self.correlationId = None
		self.reserved = None

class X224ConnectionRequest:
	def __init__(self):
		self.tpktHeader = None
		self.x224Crq = None
		self.routingToken = None
		self.cookie = None
		self.rdpNegReq = None
		self.rdpCorrelationInfo = None


################################ Server X.224 Connection Confirm PDU ############################

class RDP_NEG_RSP_FLAGS(enum.IntFlag):
	EXTENDED_CLIENT_DATA_SUPPORTED = 0x01
	DYNVC_GFX_PROTOCOL_SUPPORTED = 0x02
	NEGRSP_FLAG_RESERVED = 0x04
	RESTRICTED_ADMIN_MODE_SUPPORTED = 0x08
	REDIRECTED_AUTHENTICATION_MODE_SUPPORTED = 0x10

class RDP_NEG_RSP:
	def __init__(self):
		self.type= None
		self.flags= None
		self.length= None
		self.selectedProtocol= None

class RDP_NEG_FAILURE:
	def __init__(self):
		self.type= None
		self.flags= None
		self.length= None
		self.failureCode= None

class X224ConnectionConfirm:
	def __init__(self):
		self.tpktHeader = None
		self.x224Ccf = None
		self.rdpNegData = None

########################## Client MCS Connect Initial PDU with GCC Conference Create Request ###############
class TS_UD_HEADER:
	def __init__(self):
		self.type = None
		self.length = None

class MCSConnect:
	def __init__(self):
		self.tpktHeader = None
		self.x224Data = None
		self.mcsCi = None
		self.gccCCrq = None
		self.clientCoreData = None
		self.clientSecurityData = None
		self.clientNetworkData = None
		self.clientClusterData = None
		self.clientMonitorData = None
		self.clientMessageChannelData = None
		self.clientMultitransportChannelData = None
		self.clientMonitorExtendedData = None



import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.VNC import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession


class VNCSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.parser = VNCMessageParser(self)
		self.status = VNCSessionStatus.PROTOCOL_EXCH
		self.protocolversion = None
		self.server_challenge = None

	def __repr__(self):
		t = '== VNC Session ==\r\n'
		t += 'status:      %s\r\n' % repr(self.status)
		return t


class VNC(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.protocolversion = 'RFB 003.008\n'
		self.session.server_challenge = b'\x00' * 16

	def parse_settings(self):
		if 'protocolversion' in self.settings:
			self.session.protocolversion = self.settings['protocolversion']
			if self.session.protocolversion[-1] != '\n':
				self.session.protocolversion += '\n'
		if 'server_challenge' in self.settings:
			self.session.server_challenge = bytes.fromhex(self.settings['server_challenge'])

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# send protocolversion
		pv = ProtocolVersion()
		pv.protocolversion = self.session.protocolversion
		await asyncio.wait_for(
				self.send_data(pv.to_bytes()), timeout=1)
			
		if self.session.status == VNCSessionStatus.PROTOCOL_EXCH:
			sh = SecurityHandshake()
			sh.security_types = [VNCSecurityTypes.VNC_AUTHENTICATION]
			await self.send_data(sh.to_bytes())
			self.session.status = VNCSessionStatus.SECURITY
					
		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]

					
			if self.session.status == VNCSessionStatus.SECURITY:
				if isinstance(msg, SecurityHandshakeResponse):
					#print('VNC client selected the following hadshake: %s' % msg.security_type.name)
					if not msg.security_type:
						continue
					if msg.security_type != VNCSecurityTypes.NONE:
						#tricky user wants to get in without auth
						cred = Credential('VNC',
							fullhash="<NOT CRED> Attacker tried noauth!"
						)
						await self.logger.credential(cred)
					if msg.security_type != VNCSecurityTypes.VNC_AUTHENTICATION:
						srh = SecurityResultHandshake()
						srh.status = SecurityResultHandshakeStatus.FAILED
						srh.err_reason = 'Authentication type not supported!'
							
						await self.send_data(srh.to_bytes())
						return
		
						#raise Exception('Client returned different authentication type than what the server supports!')
					va = VNCAuthentication()
					va.challenge = self.session.server_challenge
					await self.send_data(va.to_bytes())
					self.session.status = VNCSessionStatus.AUTHENTICATION
					continue
					
			if self.session.status == VNCSessionStatus.AUTHENTICATION:
				if isinstance(msg, VNCAuthenticationResult):
					fullhash = "$vnc$%s$%s" % (self.session.server_challenge.hex(), msg.response.hex())
					cred = Credential('VNC',
						fullhash=fullhash
					)
					await self.logger.credential(cred)
						
					srh = SecurityResultHandshake()
					srh.status = SecurityResultHandshakeStatus.FAILED
					srh.err_reason = 'Password incorrect!'
							
					await self.send_data(srh.to_bytes())
					return



#https://www.ietf.org/rfc/rfc1035.txt
import enum
import io
import asyncio
import ipaddress
import socket

class DNSResponseCode(enum.Enum):
	NOERR = 0 #No error condition
	FORMATERR = 1 #Format error - The name server was  unable to interpret the query.
	SERVERERR = 2 #Server failure - The name server was unable to process this query due to a problem with the name server.
	NAMEERR = 3 #Name Error - Meaningful only for responses from an authoritative name server, this code signifies that the domain name referenced in the query does not exist.
	NOTIMPL = 4 #Not Implemented - The name server does not support the requested kind of query.
	REFUSED = 5 #Refused - The name server refuses to perform the specified operation for policy reasons.
	RESERVED6 = 6
	RESERVED7 = 7
	RESERVED8 = 8
	RESERVED9 = 9
	RESERVED10 = 10
	RESERVED11 = 11
	RESERVED12 = 12
	RESERVED13 = 13
	RESERVED14 = 14
	RESERVED15 = 15

class DNSType(enum.Enum):
	A          = 1 #a host address (IPv4)
	NS         = 2 #an authoritative name server
	MD         = 3 #a mail destination (Obsolete - use MX)
	MF         = 4 #a mail forwarder (Obsolete - use MX)
	CNAME      = 5 #the canonical name for an alias
	SOA        = 6 #marks the start of a zone of authority
	MB         = 7 #a mailbox domain name (EXPERIMENTAL)
	MG         = 8 #a mail group member (EXPERIMENTAL)
	MR         = 9 #a mail rename domain name (EXPERIMENTAL)
	NULL       = 10 #a null RR (EXPERIMENTAL)
	WKS        = 11 #a well known service description
	PTR        = 12 #a domain name pointer
	HINFO      = 13 #host information
	MINFO      = 14 #mailbox or mail list information
	MX         = 15 #mail exchange
	TXT        = 16 #text strings
	RP         = 17 #RFC 1183 	Responsible Person 	Information about the responsible person(s) for the domain. Usually an email address with the @ replaced by a .
	AFSDB      = 18 #RFC 1183 	AFS database record 	Location of database servers of an AFS cell. This record is commonly used by AFS clients to contact AFS cells outside their local domain. A subtype of this record is used by the obsolete DCE/DFS file system.
	SIG        = 24 #RFC 2535 	Signature 	Signature record used in SIG(0) (RFC 2931) and TKEY (RFC 2930).[7] RFC 3755 designated RRSIG as the replacement for SIG for use within DNSSEC.[7]
	KEY        = 25 #RFC 2535[3] and RFC 2930[4] 	Key record 	Used only for SIG(0) (RFC 2931) and TKEY (RFC 2930).[5] RFC 3445 eliminated their use for application keys and limited their use to DNSSEC.[6] RFC 3755 designates DNSKEY as the replacement within DNSSEC.[7] RFC 4025 designates IPSECKEY as the replacement for use with IPsec.[8]
	AAAA       = 28 #RFC 3596[2] 	IPv6 address record 	Returns a 128-bit IPv6 address, most commonly used to map hostnames to an IP address of the host.
	LOC        = 29 #RFC 1876 	Location record 	Specifies a geographical location associated with a domain name
	SRV        = 33 #RFC 2782 	Service locator 	Generalized service location record, used for newer protocols instead of creating protocol-specific records such as MX.
	NAPTR      = 35 #RFC 3403 	Naming Authority Pointer 	Allows regular-expression-based rewriting of domain names which can then be used as URIs, further domain names to lookups, etc.
	KX         = 36 #RFC 2230 	Key Exchanger record 	Used with some cryptographic systems (not including DNSSEC) to identify a key management agent for the associated domain-name. Note that this has nothing to do with DNS Security. It is Informational status, rather than being on the IETF standards-track. It has always had limited deployment, but is still in use.
	CERT       = 37 #RFC 4398 	Certificate record 	Stores PKIX, SPKI, PGP, etc.
	DNAME      = 39 #RFC 6672 		Alias for a name and all its subnames, unlike CNAME, which is an alias for only the exact name. Like a CNAME record, the DNS lookup will continue by retrying the lookup with the new name.
	OPT        = 41 #RFC 6891 	Option 	This is a "pseudo DNS record type" needed to support EDNS
	APL        = 42 #RFC 3123 	Address Prefix List 	Specify lists of address ranges, e.g. in CIDR format, for various address families. Experimental.
	DS         = 43 #RFC 4034 	Delegation signer 	The record used to identify the DNSSEC signing key of a delegated zone
	SSHFP      = 44 #RFC 4255 	SSH Public Key Fingerprint 	Resource record for publishing SSH public host key fingerprints in the DNS System, in order to aid in verifying the authenticity of the host. RFC 6594 defines ECC SSH keys and SHA-256 hashes. See the IANA SSHFP RR parameters registry for details.
	IPSECKEY   = 45 #RFC 4025 	IPsec Key 	Key record that can be used with IPsec
	RRSIG      = 46 #RFC 4034 	DNSSEC signature 	Signature for a DNSSEC-secured record set. Uses the same format as the SIG record.
	NSEC       = 47 #RFC 4034 	Next Secure record 	Part of DNSSEC—used to prove a name does not exist. Uses the same format as the (obsolete) NXT record.
	DNSKEY     = 48 #RFC 4034 	DNS Key record 	The key record used in DNSSEC. Uses the same format as the KEY record.
	DHCID      = 49 #RFC 4701 	DHCP identifier 	Used in conjunction with the FQDN option to DHCP
	NSEC3      = 50 #RFC 5155 	Next Secure record version 3 	An extension to DNSSEC that allows proof of nonexistence for a name without permitting zonewalking
	NSEC3PARAM = 51 #RFC 5155 	NSEC3 parameters 	Parameter record for use with NSEC3
	TLSA       = 52 #RFC 6698 	TLSA certificate association 	A record for DANE. RFC 6698 defines "The TLSA DNS resource record is used to associate a TLS server certificate or public key with the domain name where the record is found, thus forming a 'TLSA certificate association'".
	HIP        = 55 #RFC 8005 	Host Identity Protocol 	Method of separating the end-point identifier and locator roles of IP addresses.
	CDS        = 59 #RFC 7344 	Child DS 	Child copy of DS record, for transfer to parent
	CDNSKEY    = 60 #RFC 7344 	Child DNSKEY 	Child copy of DNSKEY record, for transfer to parent
	OPENPGPKEY = 61 #RFC 7929 	OpenPGP public key record 	A DNS-based Authentication of Named Entities (DANE) method for publishing and locating OpenPGP public keys in DNS for a specific email address using an OPENPGPKEY DNS resource record.
	TKEY       = 249 #RFC 2930 	Transaction Key record 	A method of providing keying material to be used with TSIG that is encrypted under the public key in an accompanying KEY RR.[10]
	TSIG       = 250 #RFC 2845 	Transaction Signature 	Can be used to authenticate dynamic updates as coming from an approved client, or to authenticate responses as coming from an approved recursive name server[11] similar to DNSSEC.
	IXFR 	   = 251 #RFC 1996 	Incremental Zone Transfer 	Requests a zone transfer of the given zone but only differences from a previous serial number. This request may be ignored and a full (AXFR) sent in response if the authoritative server is unable to fulfill the request due to configuration or lack of required deltas.
	AXFR       = 252 #RFC 1035[1] 	Authoritative Zone Transfer 	Transfer entire zone file from the master name server to secondary name servers.
	ANY        = 255 #ANYTHING
	URI        = 256 #RFC 7553 	Uniform Resource Identifier 	Can be used for publishing mappings from hostnames to URIs.
	CAA        = 257 #RFC 6844 	Certification Authority Authorization 	DNS Certification Authority Authorization, constraining acceptable CAs for a host/domain
	


class DNSClass(enum.Enum):
	IN = 1 #the Internet
	CS = 2 #the CSNET class (Obsolete - used only for examples in some obsolete RFCs)
	CH = 3 #the CHAOS class
	HS = 4 #Hesiod [Dyer 87]
	ANY = 255

class DNSOpcode(enum.Enum):
	QUERY = 0 #a standard query ()
	IQUERY = 1 #an inverse query ()
	STATUS = 2 #a server status request ()
	RESERVED3 = 3
	RESERVED4 = 4
	RESERVED5 = 5
	RESERVED6 = 6
	RESERVED7 = 7
	RESERVED8 = 8
	RESERVED9 = 9
	RESERVED10 = 10
	RESERVED11 = 11
	RESERVED12 = 12
	RESERVED13 = 13
	RESERVED14 = 14
	RESERVED15 = 15

#the upper bit is always zero and must never be set
class DNSFlags(enum.IntFlag):
	AA         = 0x0040 #Authoritative Answer
	TC         = 0x0020 #TrunCation
	RD         = 0x0010 #Recursion Desired
	RA         = 0x0008  #Recursion Available
	RESERVED1  = 0x0004
	RESERVED2  = 0x0002
	RESERVED3  = 0x0001

class DNSResponse(enum.Enum):
	REQUEST = 0 #Query
	RESPONSE = 1

class DNSPacket():
	def __init__(self, proto = socket.SOCK_STREAM):
		self.proto     = proto
		self.PACKETLEN = None #this is for TCP
		self.TransactionID = None
		self.QR = None
		self.Opcode = None
		self.FLAGS = None
		self.Rcode = None
		self.QDCOUNT = None
		self.ANCOUNT = None
		self.NSCOUNT = None
		self.ARCOUNT = None

		self.Questions = []
		self.Answers   = []
		self.Authorities = []
		self.Additionals = []

	async def from_streamreader(reader, proto = socket.SOCK_DGRAM):
		if proto == socket.SOCK_DGRAM:
			data = await reader.read()
			return DNSPacket.from_bytes(data)
		else:
			plen_bytes = await reader.readexactly(2)
			plen = int.from_bytes(plen_bytes, byteorder = 'big', signed=False)
			data = await reader.readexactly(plen)
			return DNSPacket.from_bytes(plen_bytes + data, proto = proto)


	def from_bytes(bbuff, proto = socket.SOCK_DGRAM):
		return DNSPacket.from_buffer(io.BytesIO(bbuff), proto)

	def from_buffer(buff, proto = socket.SOCK_DGRAM):
		packet = DNSPacket()
		packet.proto = proto
		if packet.proto == socket.SOCK_STREAM:
			packet.PACKETLEN = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
			buff = io.BytesIO(buff.read()) #need to repack this to a new buffer because the length field is not taken into account when calculating compressed DNSName 

		packet.TransactionID = buff.read(2)
		temp = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)

		packet.QR     = DNSResponse(temp >> 15)
		packet.Opcode = DNSOpcode((temp & 0x7800) >> 11) 
		packet.FLAGS  = DNSFlags((temp  >> 4) & 0x7F)
		packet.Rcode  = DNSResponseCode(temp & 0xF)

		packet.QDCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		packet.ANCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		packet.NSCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		packet.ARCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		
		for i in range(0, packet.QDCOUNT):
			dnsq = DNSQuestion.from_buffer(buff)
			packet.Questions.append(dnsq)

		
		for i in range(0, packet.ANCOUNT):
			dnsr = DNSResourceParser.from_buffer(buff)
			packet.Answers.append(dnsr)

		for i in range(0, packet.NSCOUNT):
			dnsr = DNSResourceParser.from_buffer(buff)
			packet.Authorities.append(dnsr)

		for i in range(0, packet.ARCOUNT):
			dnsr = DNSResourceParser.from_buffer(buff)
			packet.Additionals.append(dnsr)

		return packet
		

	def __repr__(self):
		t = '== DNS Packet ==\r\n'
		t+= 'TransactionID:  %s\r\n' % self.TransactionID.hex()
		t+= 'QR:  %s\r\n' % self.QR.name
		t+= 'Opcode: %s\r\n' % self.Opcode.name
		t+= 'FLAGS: %s\r\n' % repr(self.FLAGS)
		t+= 'Rcode: %s\r\n' % self.Rcode.name
		t+= 'QDCOUNT: %s\r\n' % self.QDCOUNT
		t+= 'ANCOUNT: %s\r\n' % self.ANCOUNT
		t+= 'NSCOUNT: %s\r\n' % self.NSCOUNT
		t+= 'ARCOUNT: %s\r\n' % self.ARCOUNT

		if len(self.Questions) > 0:
			for question in self.Questions:
				t+= repr(question)

		if len(self.Answers) > 0:
			for answer in self.Answers:
				t+= repr(answer)

		if len(self.Authorities) > 0:
			for answer in self.Authorities:
				t+= repr(answer)

		if len(self.Additionals) > 0:
			for answer in self.Additionals:
				t+= repr(answer)

		return t

	def to_bytes(self):
		t = self.TransactionID

		a  = self.Rcode.value
		a |= (self.FLAGS << 4 ) & 0x7F0
		a |= (self.Opcode.value << 11) & 0x7800
		a |= (self.QR.value << 15) & 0x8000
		t += a.to_bytes(2, byteorder = 'big', signed = False)

		t += self.QDCOUNT.to_bytes(2, byteorder = 'big', signed=False)
		t += self.ANCOUNT.to_bytes(2, byteorder = 'big', signed=False)
		t += self.NSCOUNT.to_bytes(2, byteorder = 'big', signed=False)
		t += self.ARCOUNT.to_bytes(2, byteorder = 'big', signed=False)


		for q in self.Questions:
			t += q.to_bytes()

		for q in self.Answers:
			t += q.to_bytes()

		for q in self.Authorities:
			t += q.to_bytes()

		for q in self.Additionals:
			t += q.to_bytes()
		
		if self.proto == socket.SOCK_STREAM:
			self.PACKETLEN = len(t)
			t = self.PACKETLEN.to_bytes(2, byteorder = 'big', signed=False) + t

		return t

	def construct(TID, response,  flags = 0, opcode = DNSOpcode.QUERY, rcode = DNSResponseCode.NOERR, 
					questions= [], answers= [], authorities = [], additionals = [], proto = socket.SOCK_DGRAM):
		packet = DNSPacket()
		packet.proto   = proto
		packet.TransactionID = TID
		packet.QR      = response
		packet.Opcode  = opcode
		packet.FLAGS   = flags
		packet.Rcode   = rcode
		packet.QDCOUNT = len(questions)
		packet.ANCOUNT = len(answers)
		packet.NSCOUNT = len(authorities)
		packet.ARCOUNT = len(additionals)

		packet.Questions   = questions
		packet.Answers     = answers
		packet.Authorities = authorities
		packet.Additionals = additionals

		return packet


class DNSQuestion():
	def __init__(self):
		self.QNAME  = None
		self.QTYPE  = None
		self.QCLASS = None
		self.QU     = None

	def from_bytes(bbuff):
		return DNSQuestion.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		qst = DNSQuestion()
		qst.QNAME  = DNSName.from_buffer(buff)
		qst.QTYPE  = DNSType(int.from_bytes(buff.read(2), byteorder = 'big', signed = False))
		temp = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		qst.QCLASS = DNSClass(temp & 0x7fff)
		qst.QU     = bool((temp & 0x8000) >> 15)

		return qst

	def to_bytes(self):
		t  = self.QNAME.to_bytes()
		t += self.QTYPE.value.to_bytes(2, byteorder = 'big', signed = False)
		a  = self.QCLASS.value
		a |= int(self.QU) << 15
		t += a.to_bytes(2, byteorder = 'big', signed = False)

		return t

	def construct(qname, qtype, qclass, qu = False):
		qst = DNSQuestion()
		qst.QNAME     = DNSName.construct(qname)
		qst.QTYPE     = qtype
		qst.QCLASS    = qclass
		qst.QU        = qu
		return qst

	def __repr__(self):
		t = '== DNSQuestion ==\r\n'
		t+= 'QNAME:  %s\r\n' % self.QNAME.name
		t+= 'QTYPE:  %s\r\n' % self.QTYPE.name
		t+= 'QCLASS: %s\r\n' % self.QCLASS.name
		t+= 'QU    : %s\r\n' % self.QU
		return t

class DNSOPT():
	def __init__(self):
		self.Code   = None
		self.Length = None
		self.Value  = None

		#temp variable!
		self.size = None

	def from_bytes(buff):
		return DNSOPT.from_buffer(io.BytesIO(buff))

	def from_buffer(buff):
		o = DNSOPT()
		o.Code = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		o.Length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		o.Value = buff.read(o.Length)
		o.size = o.Length + 2
		return o


class DNSOPTResource():
	#https://tools.ietf.org/html/rfc6891 Section 6.1.2.
	#Comments from the author: <REDACTED PROFANITIES> Seriosuly why did you re-invent the format????
	def __init__(self):
		self.NAME     = None
		self.TYPE     = None
		self.UDPSIZE  = None
		self.EXTRCODE = None
		self.VERSION  = None
		self.DO       = None
		self.Z        = None
		self.RDLENGTH = None
		self.RDATA    = None

		self.options = []

	def from_bytes(bbuff):
		return DNSOPTResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		rsc = DNSOPTResource()
		rsc.NAME     = DNSName.from_buffer(buff)
		rsc.TYPE     = DNSType(int.from_bytes(buff.read(2), byteorder = 'big', signed = False))
		rsc.UDPSIZE  = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		rsc.EXTRCODE = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		rsc.VERSION  = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		temp = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		rsc.DO       = bool(temp & 0x8000)
		rsc.Z        = temp & 0x7fff
		rsc.RDLENGTH = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)

		rsc.RDATA    = buff.read(rsc.RDLENGTH)

		#TODO Further parsing the opt elements
		"""
		i = rsc.RDLENGTH
		if rsc.RDLENGTH > 0:
			opt = DNSOPT.from_bytes(rsc.RDATA)
			rsc.options.append(opt)
		"""
		return rsc

	def __repr__(self):
		t = '== DNSOPTResource ==\r\n'
		t+= 'NAME:  %s\r\n' % self.NAME.name
		t+= 'TYPE:  %s\r\n' % self.TYPE.name
		t+= 'UDPSIZE : %d\r\n' % self.UDPSIZE
		t+= 'EXTRCODE: %s\r\n' % self.EXTRCODE
		t+= 'VERSION: %s\r\n' % self.VERSION
		t+= 'DO: %s\r\n' % repr(self.DO)
		t+= 'RDLENGTH: %d\r\n' % self.RDLENGTH
		t+= 'RDATA: %s\r\n' % repr(self.RDATA)
		return t

	def to_bytes(self):
		t  = self.NAME.to_bytes()
		t += self.TYPE.value.to_bytes(2, byteorder = 'big', signed = False)
		t += self.UDPSIZE.to_bytes(2, byteorder = 'big', signed = False)
		t += self.EXTRCODE.to_bytes(1, byteorder = 'big', signed = False)
		t += self.VERSION.to_bytes(1, byteorder = 'big', signed = False)

		a = self.Z
		a |= int(self.DO) << 15
		t += a.to_bytes(2, byteorder = 'big', signed = False)
		t += self.RDLENGTH.to_bytes(2, byteorder = 'big', signed = False)
		t += self.RDATA

		return t

class DNSResource():
	def __init__(self):
		self.NAME     = None
		self.TYPE     = None
		self.CLASS    = None
		self.CFLUSH   = None
		self.TTL      = None
		self.RDLENGTH = None
		self.RDATA    = None

	#this method will parse the Resource object until RDATA, which will be parsed by the object inheriting it
	def parse_header(self, buff):
		self.NAME     = DNSName.from_buffer(buff)
		self.TYPE     = DNSType(int.from_bytes(buff.read(2), byteorder = 'big', signed = False))
		temp = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		self.CLASS    = DNSClass(temp & 0x7fff)
		self.CFLUSH   = bool((temp & 0x8000) >> 15)
		self.TTL      = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		self.RDLENGTH = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		#this is here to support proxying of packets
		pos = buff.tell()
		self.RDATA    = buff.read(self.RDLENGTH)
		buff.seek(pos, io.SEEK_SET)
		return

	def from_bytes(bbuff):
		return DNSResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		res = DNSResource()
		res.parse_header(buff)
		res.RDATA = buff.read(res.RDLENGTH)
		return res
	
	def to_bytes(self):
		t  = self.NAME.to_bytes()
		t += self.TYPE.value.to_bytes(2, byteorder = 'big', signed = False)
		a  = self.CLASS.value
		a |= int(self.CFLUSH) << 15
		t += a.to_bytes(2, byteorder = 'big', signed = False)
		t += self.TTL.to_bytes(4, byteorder = 'big', signed = False)
		t += self.RDLENGTH.to_bytes(2, byteorder = 'big', signed = False)
		t += self.RDATA

		return t

	def construct(self, rname, rtype, rdata, ttl = 30, rclass = DNSClass.IN, cflush = False):
		res = DNSResource()
		res.NAME     = DNSName.construct(rname)
		res.TYPE     = rtype
		res.CLASS    = rclass
		res.CFLUSH   = cflush
		res.TTL      = ttl
		res.RDATA    = rdata
		return res


	def __repr__(self):
		t = '== DNSResource ==\r\n'
		t+= 'NAME:  %s\r\n' % self.NAME.name
		t+= 'TYPE:  %s\r\n' % self.TYPE.name
		t+= 'CLASS : %s\r\n' % self.CLASS.name
		t+= 'CFLUSH: %s\r\n' % self.CFLUSH
		t+= 'TTL: %s\r\n' % self.TTL
		t+= 'RDLENGTH: %s\r\n' % self.RDLENGTH
		t+= 'RDATA: %s\r\n' % repr(self.RDATA)
		return t
		
#the logic here is: all resource types contain different data in the RDATA filed
#we describe the custom data for each and every different type
#and define a new object inheriting for the base object 
class DNSAResource(DNSResource):
	def __init__(self):
		DNSResource.__init__(self)
		self.ipaddress = None

	def from_bytes(bbuff):
		return DNSAResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		res = DNSAResource()
		res.parse_header(buff)
		res.ipaddress = ipaddress.IPv4Address(buff.read(res.RDLENGTH))
		return res

	def construct(rname, ipv4, ttl = 3000, rclass = DNSClass.IN, cflush = False):
		res = DNSAResource()
		res.NAME     = DNSName.construct(rname)
		res.TYPE     = DNSType.A
		res.CLASS    = rclass
		res.CFLUSH   = cflush
		res.TTL      = ttl
		res.ipaddress = ipv4
		
		res.RDATA    = res.ipaddress.packed
		res.RDLENGTH = len(res.RDATA) #should be 4

		return res

	def __repr__(self):
		t = '=== DNS A ===\r\n'
		t += DNSResource.__repr__(self)
		t += 'IP address: %s\r\n' % str(self.ipaddress)
		return t


class DNSAAAAResource(DNSResource):
	def __init__(self):	
		DNSResource.__init__(self)
		self.ipaddress = None

	def from_bytes(bbuff):
		return DNSAAAAResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		res = DNSAAAAResource()
		res.parse_header(buff)
		res.ipaddress = ipaddress.IPv6Address(buff.read(res.RDLENGTH))
		return res

	def construct(rname, ipv6, ttl = 3000, rclass = DNSClass.IN, cflush = False):
		res = DNSAAAAResource()
		res.NAME     = DNSName.construct(rname)
		res.TYPE     = DNSType.AAAA
		res.CLASS    = rclass
		res.CFLUSH   = cflush
		res.TTL      = ttl
		res.ipaddress = ipv6
		
		res.RDATA    = res.ipaddress.packed
		res.RDLENGTH = len(res.RDATA) #should be 16

		return res

	def __repr__(self):
		t = '=== DNS AAAA ===\r\n'
		t += DNSResource.__repr__(self)
		t += 'IP address: %s\r\n' % str(self.ipaddress)
		return t

class DNSPTRResource(DNSResource):
	def __init__(self):	
		DNSResource.__init__(self)
		self.domainname = None

	def from_bytes(bbuff):
		return DNSPTRResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		res = DNSPTRResource()
		res.parse_header(buff)
		res.domainname = DNSName.from_buffer(buff)
		return res

	def construct(rname, domainname, ttl = 3000, rclass = DNSClass.IN, cflush = False):
		res = DNSPTRResource()
		res.NAME     = DNSName.construct(rname)
		res.TYPE     = DNSType.PTR
		res.CLASS    = rclass
		res.CFLUSH   = cflush
		res.TTL      = ttl
		res.domainname = domainname
		
		res.RDATA    = DNSName.construct(res.domainname)
		res.RDLENGTH = len(res.RDATA) #should be 16

		return res

	def __repr__(self):
		t = '=== DNS PTR ===\r\n'
		t += DNSResource.__repr__(self)
		t += 'PTR name: %s\r\n' % str(self.domainname)
		return t

class DNSSRVResource(DNSResource):
	def __init__(self):	
		DNSResource.__init__(self)
		self.Service = None
		self.Proto = None
		self.Name = None
		self.Priority = None
		self.Weight = None
		self.Port = None
		self.Target = None

		#_Service._Proto.Name TTL Class SRV Priority Weight Port Target
		
	def from_bytes(bbuff):
		return DNSSRVResource.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		res = DNSSRVResource()
		res.parse_header(buff)
		try:
			if res.NAME.name.count('.') == 4:
				res.Service, res.Proto, res.Name, tld = res.NAME.name.split('.')
				print(res.Service, res.Proto, res.Name, tld)
			else:
				res.Service = res.NAME.name
		except Exception as e:
			print(res.NAME.name)
			raise(e)
		res.Priority = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		res.Weight = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		res.Port = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		res.Target = DNSName.from_buffer(buff)

		return res

	def __repr__(self):
		t = '=== DNS SRV ===\r\n'
		t += DNSResource.__repr__(self)
		t += 'Priority: %s\r\n' % str(self.Priority)
		t += 'Weight: %s\r\n' % str(self.Weight)
		t += 'Port: %s\r\n' % str(self.Port)
		t += 'Target: %s\r\n' % str(self.Target)
		return t

class DNSResourceParser():
	def from_bytes(bbuff):
		return DNSResourceParser.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		pos = buff.tell()

		resname = DNSName.from_buffer(buff)
		restype = DNSType(int.from_bytes(buff.read(2), byteorder = 'big', signed = False))

		#Extended dns?
		if restype == DNSType.OPT:
			buff.seek(pos, io.SEEK_SET)
			return DNSOPTResource.from_buffer(buff)

		#nope.
		else:
			#rewinding the buffer
			buff.seek(pos, io.SEEK_SET)
			#now parse for various types
			if restype == DNSType.A:
				rsc = DNSAResource.from_buffer(buff)
			elif restype == DNSType.AAAA:
				rsc = DNSAAAAResource.from_buffer(buff)
			elif restype == DNSType.PTR:
				rsc = DNSPTRResource.from_buffer(buff)
			elif restype == DNSType.SRV:
				rsc = DNSSRVResource.from_buffer(buff)

			#catch-all for not fully implemented or unknown types
			#feel free to expand the if-else statement above...
			else:
				rsc = DNSResource.from_buffer(buff)
				

		return rsc


class DNSName():
	def __init__(self):
		self.name            = ''
		#variables below are for parsing only
		self.compressed      = False
		self.compressed_pos  = None
		self.compressed_done = False
			
	def from_bytes(bbuff):
		return DNSName.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		dnsname = DNSName()
		dnsname.parse(buff)
		return dnsname
		

	def parse(self, data, rec = False):
		#this code is ugly :(
		if self.compressed_done:
			return
		temp = data.read(1)[0]
		if not self.compressed:
			self.compressed = (temp & 0xC0) >> 6 == 0x3
			if self.compressed:
				data.seek(-1, io.SEEK_CUR)
				self.compressed_pos = data.tell()
				temp = int.from_bytes(data.read(2), byteorder = 'big', signed = False)
				ptr = temp & 0x3fff
				data.seek(ptr, io.SEEK_SET)
				self.parse(data, rec)
				data.seek(self.compressed_pos + 2, io.SEEK_SET)
				self.compressed_done = True
				return
		
		length = temp
		if length == 0:			
			return

		if length < 63:
			if rec:
				self.name += '.' + data.read(length).decode()
			else:
				self.name += data.read(length).decode()
			self.parse(data, True)

	def construct(name):
		dnsname = DNSName()
		dnsname.name = name
		return dnsname


	def to_bytes(self):
		#not using compression here! implement yourself and do a PR please
		#will give you chokolate
		t = b''
		for label in self.name.split('.'):
			t += len(label).to_bytes(1, byteorder = 'big', signed = False)
			t += label.encode()

		t += b'\x00'
		return t

	def __repr__(self):
		return self.name

	"""
	def decode_NS(self, encoded_name):
		#encoded http://www.ietf.org/rfc/rfc1001.txt
		transform = [((i - 0x41)& 0x0F) for i in encoded_name]
		i = 0
		while i < len(transform):
			self.QNAME += chr(transform[i] << 4 | transform[i+1] ) 
			i+=2
	"""


#!/usr/bin/env python
import os
import sys
import enum
import json
import atexit
import datetime
import socket
import platform
import traceback
import ipaddress


from responder3.crypto.hashing import *
import asyncio


class ConnectionStatus(enum.Enum):
	OPENED = 0
	CLOSED = 1
	STATELESS = 3


class ConnectionFactory:
	def __init__(self, rdnsd, rdns_resolver):
		"""
		Creates Connetion object from the socket input.
		:param rdnsd: shared dictionary to speed up the rdns lookup
		:type rdnsd: dict created via multiprocessing.Manager()
		"""
		self.rdnsd = rdnsd
		self.resolver = rdns_resolver
		self.connection_id = 0
		

	async def from_streamwriter(self, writer):
		"""
		Creates Connection object from streamwriter
		:param writer: Streamwriter
		:type writer: asyncio.Streamwriter.
		:return: responder3.core.commons.Connection.
		"""
		protocoltype = writer.get_extra_info('socket').getsockopt(socket.SOL_SOCKET, socket.SO_TYPE)
		cid = self.connection_id
		self.connection_id += 1
		con = Connection()
		con.writer = writer
		con.connection_id = cid
		con.local_ip, con.local_port   = writer.get_extra_info('sockname')
		con.remote_ip, con.remote_port = writer.get_extra_info('peername')
		
		await self.lookup_rdns(con)
		return con
		
	async def lookup_rdns(self, con):
		"""
		Resolves the remote host's IP address to a DNS address.
		First checks if the address has already been resolved by looking it up in the shared rdns dictionary
		:param con: The Connection object specifies the connection settings
		:type con: Connection
		:return: Nothing
		"""
		
		"""
		#failback for rdns resolution
		#this will work, but make the whole app slow
		
		if con.remote_ip in self.rdnsd:
			con.remote_dns = self.rdnsd[con.remote_ip]
		
		else:
			try:
				con.remote_dns = socket.gethostbyaddr(con.remote_ip)[0]
			except Exception as e:
				pass

			self.rdnsd[con.remote_ip] = con.remote_dns
		"""
		
		if con.remote_ip in self.rdnsd:
			con.remote_dns = self.rdnsd[con.remote_ip]
			
		else:
			con.remote_dns = await self.resolver.resolve(con.remote_ip)
			print(con.remote_dns)
			self.rdnsd[con.remote_ip] = con.remote_dns
		

		
class Connection:
	def __init__(self):
		"""
		Keeps all the connection related information that is used for logging and/or connection purposes
		"""
		self.connection_id = None
		self.remote_ip   = None
		self.remote_dns  = None
		self.remote_port = None
		self.local_ip    = None
		self.local_port  = None
		self.timestamp   = datetime.datetime.utcnow()
		self.writer = None

	def get_remote_print_address(self):
		"""
		Returns the remote peer's address in a printable format
		:return: str
		"""
		return '%s:%d' % (str(self.remote_ip), int(self.remote_port))

	def get_local_print_address(self):
		"""
		Returns the local address in a printable format
		:return: str
		"""
		return '%s:%d' % (str(self.local_ip), int(self.local_port))

	def get_remote_address(self):
		"""
		Returns the remote peer's address in a socket friendly format
		:return: tuple
		"""
		return str(self.remote_ip), int(self.remote_port)

	def get_local_address(self):
		"""
		Returns the local address in a socket friendly format
		:return: tuple
		"""
		return str(self.local_ip), int(self.local_port)

	def to_dict(self):
		"""
		Converts the object to a dict
		:return: dict
		"""
		t = {}
		t['connection_id'] = self.connection_id
		t['remote_ip']   = self.remote_ip
		t['remote_port'] = self.remote_port
		t['remote_dns']  = self.remote_dns
		t['local_ip']    = self.local_ip
		t['local_port']  = self.local_port
		t['timestamp']   = self.timestamp
		return t

	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
		
	@staticmethod
	def from_dict(d):
		c = Connection()
		c.connection_id = d['connection_id']
		c.remote_ip   = d['remote_ip']
		c.remote_dns  = d['remote_dns']
		c.remote_port = d['remote_port']
		c.local_ip    = d['local_ip']
		c.local_port  = d['local_port']
		c.timestamp   = isoformat2dt(d['timestamp'])
		return c
		
		
	@staticmethod
	def from_json(data):
		return Connection.from_dict(json.loads(data))

	def __repr__(self):
		return str(self)

	def __str__(self):
		if self.remote_dns is not None and self.remote_dns != 'NA':
			return '[%s] %s:%d -> %s' % (self.timestamp.isoformat(), self.remote_dns, self.remote_port, self.get_local_print_address())
		else:
			return '[%s] %s-> %s' % (self.timestamp.isoformat(), self.get_remote_print_address(), self.get_local_print_address())


class UniversalEncoder(json.JSONEncoder):
	"""
	Used to override the default json encoder to provide a direct serialization for formats
	that the default json encoder is incapable to serialize
	"""
	def default(self, obj):
		if isinstance(obj, datetime.datetime):
			return obj.isoformat()
		elif isinstance(obj, enum.Enum):
			return obj.value
		elif isinstance(obj, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
			return str(obj)
		elif hasattr(obj, 'to_dict'):
			return obj.to_dict()
		else:
			return json.JSONEncoder.default(self, obj)


def timestamp2datetime(dt):
	"""
	Converting Windows timestamps to datetime.datetime format
	:param dt: Windows timestamp as array of bytes
	:type dt: bytearray
	:return: datetime.datetime
	"""
	us = int.from_bytes(dt, byteorder='little')/ 10.
	return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=us)


class PoisonerMode(enum.Enum):
	"""
	Enum to specify the posioner module's mode of operation
	"""
	SPOOF = 'SPOOF'
	ANALYSE = 'ANALYSE'


class ServerFunctionality(enum.Enum):
	"""
	Enum to specify the server's mode of operation
	"""
	HONEYPOT = 0
	SERVER   = 1
	TARPIT   = 2


# values MUST be lists!
defaultports = {
	"DNS"  : [(53, 'udp'),(53, 'tcp')],
	"DHCP" : [(67, 'udp')],
	"NTP"  : [(123, 'udp')],
	"SSH"  : [(22, 'tcp')],
	"TELNET"  : [(23, 'tcp')],
	"HTTP" : [(80, 'tcp')],
	"KERBEROS": [(88, 'tcp')],
	"HTTPS": [(443, 'tcp')],
	"FTP"  : [(21, 'tcp')],
	"SMTP" : [(25, 'tcp')],
	"POP3" : [(110, 'tcp')],
	"POP3S": [(995, 'tcp')],
	"IMAP" : [(143, 'tcp')],
	"IMAPS": [(993, 'tcp')],
	"SMB"  : [(445, 'tcp')],
	"LDAP" : [(389, 'tcp')],
	"RLOGIN" : [(513, 'tcp')],
	"LDAPS" : [(636, 'tcp')],
	"NBTNS": [(137, 'udp')],
	"SOCKS5":[(1080, 'tcp')],
	"MSSQL":[(1433, 'tcp')],
	"TNS":[(1521, 'tcp')],
	"MYSQL":[(3306, 'tcp')],
	"VNC":[(5900, 'tcp')],
	"SIP":[(5060, 'tcp')], #UDP maybe?
	"SIPS":[(5061, 'tcp')],
	"LLMNR": [(5355, 'udp')],
	"MDNS" : [(5353, 'udp')],
	"HTTPProxy":[(8080, 'tcp')],
	"R3M":[(55551,'tcp')]
}


def byealex(name_of_pid):
	pidfile = str(name_of_pid)
	os.remove(pidfile)


def handle_systemd(pidfile):
	if os.path.isfile(pidfile):
		print ("%s already exists, exiting" % pidfile)
		sys.exit()

	pid = str(os.getpid())
	with open(pidfile, 'w') as f:
		f.write(pid)
	
	atexit.register(byealex,pidfile)


class ResponderPlatform(enum.Enum):
	"""
	Enum to specify the platform the code is running on
	"""
	UNKNOWN = 0
	WINDOWS = 1
	LINUX   = 2
	MAC     = 3


def get_platform():
	"""
	Detects the current platform
	:return: ResponderPlatform
	"""
	p = platform.system()
	if p == 'Linux':
		return ResponderPlatform.LINUX
	elif p == 'Windows':
		return ResponderPlatform.WINDOWS
	elif p == 'Darwin':
		return ResponderPlatform.MAC
	else:
		return ResponderPlatform.UNKNOWN


# thank you Python developers who after A FUCKING DECADE
# could not figure out a way to make your datetime.datetime
# object in a format that is reversible
# now it's either this bullshit "solution" OR installing a 3rd party
# lib that have to GUESS YOUR SHITTY FORMAT
# PS: "isoformat" doesn't even conforming to the ISO standard..
def isoformat2dt(isostr):
	"""
	Converts back the string result of datetime.datateime.isoformat() to a datetime.datetime object
	:param isostr: string output of datetime.datetime.isoformat()
	:type isostr: str
	:return: datetime.datetime
	"""
	dt, _, us = isostr.partition(".")
	dt = datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S")
	us = int(us.rstrip("Z"), 10)
	return dt + datetime.timedelta(microseconds=us)


# https://gist.github.com/ImmortalPC/c340564823f283fe530b
def hexdump(src, length=16, sep='.'):
	"""
	Pretty printing binary data blobs
	:param src: Binary blob
	:type src: bytearray
	:param length: Size of data in each row
	:type length: int
	:param sep: Character to print when data byte is non-printable ASCII
	:type sep: str(char)
	:return: str
	"""
	result = []

	for i in range(0, len(src), length):
		subSrc = src[i:i+length]
		hexa = ''
		isMiddle = False
		for h in range(0,len(subSrc)):
			if h == length/2:
				hexa += ' '
			h = subSrc[h]
			if not isinstance(h, int):
				h = ord(h)
			h = hex(h).replace('0x', '')
			if len(h) == 1:
				h = '0'+h
			hexa += h+' '
		hexa = hexa.strip(' ')
		text = ''
		for c in subSrc:
			if not isinstance(c, int):
				c = ord(c)
			if 0x20 <= c < 0x7F:
				text += chr(c)
			else:
				text += sep
		result.append(('%08X:  %-'+str(length*(2+1)+1)+'s  |%s|') % (i, hexa, text))

	return '\n'.join(result)


def get_mutual_preference(preference, offered):
	# this is a commonly used algo when we need to determine the mutual option
	# which is both supported by the client and the server, in order of the
	# server's preference
	"""
	Generic function to determine which option to use from two lists of options offered by two parties.
	Returns the option that is mutual and in the highes priority of the preference
	:param preference: A list of options where the preference is set by the option's position in the list (lower is most preferred)
	:type preference: list
	:param offered: A list of options that the other party can offer
	:type offered: list
	:return: tuple
	"""
	clinet_supp = set(offered)
	server_supp = set(preference)
	common_supp = server_supp.intersection(clinet_supp)
	if common_supp is None:
		return None, None
	
	preferred_opt = None
	for srv_option in preference:
		for common_option in common_supp:
			if common_option == srv_option:
				preferred_opt = srv_option
				break
		else:
			continue
		break
	
	# getting index of the preferred option...
	preferred_opt_idx = 0
	for option in offered:
		if option == preferred_opt:
			# preferred_dialect_idx += 1
			break
		preferred_opt_idx += 1

	return preferred_opt, preferred_opt_idx


def read_element(line, marker = ' ', marker_end = None, toend = False):
	"""
	Helper function to read a substring from a string.
	Returns a tuple containing the substring and the remaining data starting from the end of the substring.
	:param line: the data to read from
	:type line: str
	:param marker: if marker_end is not defined it marks the end of the returned substring
					if marker is defined it marks the beginning of the returned substring
	:type marker: str
	:param marker_end: if defined it marks the end of the returned substring
	:param marker_end: str
	:param toend:
	:type toend: bool
	:return: tuple
	"""
	if marker_end is None:
		m = line.find(marker)
		if m == -1:
			if toend:
				return line, ''
			print(line)
			raise Exception('Marker not found!')
		element = line[:m]
		line = line[m+len(marker):]
		return element, line
	else:
		start = line.find(marker)
		end   = line.find(marker_end)
		if start == -1 or end == -1 or end <= start:
			raise Exception('Marker not found!')
		element = line[start:end]
		line = line[end:]
		return element, line


def tracefunc(frame, event, arg, indent=[0]):
	if event == "call":
		indent[0] += 2
		print("-" * indent[0] + "> call function", frame.f_code.co_name)
	elif event == "return":
		print("<" + "-" * indent[0], "exit function", frame.f_code.co_name)
		indent[0] -= 2
	return tracefunc
	



# https://en.wikipedia.org/wiki/Digest_access_authentication
import enum
import hashlib
import collections

from responder3.protocols.authentication.common import *
from responder3.core.logging.log_objects import Credential

def n2e(x):
	return '' if x is None else x

class DIGESTAlgo(enum.Enum):
	MD5 = 'MD5'
	MD5_SESS = 'MD5-Sess'
	TOKEN = 'TOKEN'

class DIGESTStatus(enum.Enum):
	START = enum.auto()
	AUTHENTICATE = enum.auto()
	AUTHORIZE = enum.auto()
	FINISH = enum.auto()

class DIGESTAuthenticate:
	def __init__(self):		
		self.method = 'DIGEST'
		self.challenge = None
		self.realm = None
		self.domain = None
		self.nonce = None
		self.opaque = None
		self.stale = None
		self.algorithm = None
		self.qop = None #can be one of these: "auth" | "auth-int" | token
		self.auth = []
		self.uri = None

	def to_line(self):
		data = ''
		t = collections.OrderedDict()
		t['realm'] = '"%s"' % self.realm if self.realm else '""'
		if not self.nonce:
			raise Exception('DIGEST nonce needs to be set!')
		t['nonce'] = '"%s"' % self.nonce
		t['opaque'] = '"%s"' % self.opaque if self.opaque else '"AAAAAAAAAAAAAA"'
		t['stale'] = str(self.stale)
		t['algorithm'] = self.algorithm.name
		if self.domain:
			t['domain'] = self.domain
		if self.qop:
			t['qop'] = self.qop
		if self.auth:
			t['auth'] = self.auth

		tt = []
		for key in t:
			tt.append('%s=%s' % (key, t[key]))

		data += ','.join(tt)
		return data

	@staticmethod
	def from_dict(kv):
		ds = DIGESTAuthenticate()
		ds.challenge = kv['challenge']
		ds.realm = kv['realm']
		ds.domain = kv['domain']
		ds.nonce = kv['nonce']
		ds.opaque = kv['opaque']
		ds.stale = bool(kv['stale'])
		ds.algorithm = DIGESTAlgo(kv['algorithm'].upper())
		if 'domain' in kv:
			ds.domain = kv['domain']
		if 'qop' in kv:
			ds.qop_options = kv['qop'].split(',')
		if 'auth' in kv:
			ds.auth = kv['auth'].split(',')

		return ds

	@staticmethod
	def from_line(data):
		m = data.find(' ')
		t_method = data[:m].upper()
		if t_method != 'DIGEST':
			raise Exception('Expected DIGEST, got %s' % t_method)
		#method = t_method
		t_data = data[m+1:]
		kv = {}
		for elem in t_data.split(','):
			key, value = elem.split('=')
			if value[0] == '"':
				value = value[1:-1]
			kv[key] = value
		da = DIGESTAuthenticate.from_dict(kv)
		return da

class DIGESTAuthorize:
	def __init__(self):
		self.username = None#quoted
		self.realm = None
		self.uri = None
		self.qop = None
		self.nonce = None
		self.cnonce = None
		self.nonce_count = None #8 hex digits (4bytes encoded) #This MUST be specified if a qop directive is sent
		self.response = None
		self.algorithm = None
		self.opaque = None
		self.auth = None

	# Digest username="alma", realm="TEST", nonce="AAAAAAAAAAAAAAAAAAAA", uri="/", algorithm=MD5, response="d796a7170108b6b7237682d1f5baabb9", opaque="AAAAAAAAAAAAAAAAAAAA"

	def to_line(self):
		data = ''
		return data

	@staticmethod
	def from_dict(kv):
		ds = DIGESTAuthorize()
		ds.username = kv['username']
		ds.realm = kv['realm']
		ds.uri = kv['uri']
		ds.nonce = kv['nonce']
		ds.opaque = kv.get('opaque', None) #old clients discard this setting...
		ds.algorithm = DIGESTAlgo(kv['algorithm'].upper())
		ds.response = kv['response']
		if 'qop' in kv:
			ds.cnonce = kv['cnonce']
			ds.nonce_count = kv['nonce-count']
			ds.qop = kv['qop'].split(',')
		if 'auth' in kv:
			ds.auth = kv['auth'].split(',')

		return ds

	@staticmethod
	def from_line(data):
		kv = {}
		for elem in data.split(','):
			elem = elem.strip()
			key, value = elem.split('=')
			if value[0] == '"':
				value = value[1:-1]
			kv[key] = value
		da = DIGESTAuthorize.from_dict(kv)
		return da

class DIGEST:
	def __init__(self, credentials):
		self.mode = None #AUTHModuleMode SERVER or CLIENT
		self.status = DIGESTStatus.START
		self.credentials = credentials
		self.authenticate = None
		self.authorize = None

		#intermediate values
		self.hash_obj = None
		self.HA1 = None
		self.HA2 = None
		self.response = None

	def setup_defaults(self):
		self.mode = AUTHModuleMode.SERVER
		kv = {
			'challenge' : 'A'*20,
			'realm': 'TEST',
			'domain': 'TEST',
			'nonce': 'A'*20,
			'opaque': 'A'*20,
			'stale': False,
			'algorithm' : 'MD5',
			#'qop' : None,
			#'auth' : None,
		}
		self.authenticate = DIGESTAuthenticate.from_dict(kv)
		return

	def setup(self, settings):
		raise Exception('Not implemented!')

	def get_a1(self, password, ha1 = None):
		a1 = '%s:%s:%s' % (self.authorize.username, self.authorize.realm, password)
		if DIGESTAlgo.MD5_SESS:
			if not ha1:
				ha1 = self.hash_obj(a1.encode()).digest()
			a1 = b'%s:%s:%s' % (ha1, self.authorize.nonce, self.authorize.cnonce)

		return a1

	def get_a2(self, method, body_data = None):
		a2 = b'%s:%s' % (method, self.authorize.uri.encode())
		if self.authenticate.qop == 'auth-int': 
			a2 = b'%s:%s' % (a2, self.hash_obj(body_data).digest())
		return a2

	def verify_creds(self, method = None, body_data = None):
		"""
		Verifyies user creds, returns a tuple with (verification_result, credential)
		currently it products SIP hash, sorry
		also it doesnt variy creds, todo
		"""
		#print(self.authorize.username)
		#print(n2e(self.authorize.username))
		fullhash = '$sip$*%s' % ('*'.join([
											n2e(self.authenticate.uri),  #[URI_SERVER]
											n2e(self.authorize.uri),     #[URI_CLIENT]
											n2e(self.authorize.username),#[USERNAME]
											n2e(self.authorize.realm),   #[REALM]
											n2e(method),                 #[METHOD]
											n2e(''), #[URI_PREFIX]
											n2e(''), #[URI_RESOURCE]
											n2e(''), #[URI_SUFFIX]
											n2e(self.authenticate.nonce),#[NONCE_SERVER]
											n2e(self.authorize.cnonce),  #[NONCE_CLIENT]
											n2e(self.authorize.nonce_count), #[NONCE_COUNT]
											n2e(self.authorize.qop),     #[QOP]
											n2e(self.authenticate.algorithm.name), #DIRECTIVE
											n2e(self.authorize.response) #[MD5]
										]
									)
								)
		credential = Credential(
				'DIGEST',
				domain = self.authorize.realm,
				username = self.authorize.username,
				fullhash = fullhash
			)
		return AuthResult.FAIL, credential

	def do_auth(self, data = None, method = None, body_data = None):
		# body_data : is qop == 'auth-int' then digest value incorporates the body data's hash
		if self.mode == AUTHModuleMode.SERVER:
			if self.status == DIGESTStatus.START:
				if data is None:
					self.status = DIGESTStatus.AUTHORIZE
					#returning WWW-Authenticate data
					return AuthResult.CONTINUE, self.authenticate.to_line()
					
				else:
					self.status = DIGESTStatus.AUTHORIZE
					#raise Exception('DIGEST AUTH: Unexpected data in SERVER|START state')

			if self.status == DIGESTStatus.AUTHORIZE:
				if data is not None:
					#evaulation client's credentials
					self.status = DIGESTStatus.FINISH
					self.authorize = DIGESTAuthorize.from_line(data)
					return self.verify_creds(method = None, body_data = None)
				else:
					raise Exception('DIGEST AUTH: input data expected in SERVER|AUTHORIZE state')

			else:
				raise Exception('DIGEST AUTH: Unexpected SERVER state')

		elif self.mode == AUTHModuleMode.CLIENT:
			if self.status == DIGESTStatus.START:
				if data is not None:
					#returning WWW-Authorize data
					self.status = DIGESTStatus.FINISH
					self.authenticate = DIGESTAuthenticate.to_line()
					self.authorize = DIGESTAuthorize.construct(self.authenticate, credential)
					return AuthResult.OK, self.authorize.to_line()
				else:
					raise Exception('DIGEST AUTH: input data expected in SERVER|AUTHORIZE state')

		else:
			raise Exception('DIGEST Unknown mode: %s' % self.mode)


#https://tools.ietf.org/html/rfc4271
import enum
import io

class BGPMessageType(enum.Enum):
	OPEN = 1 
	UPDATE = 2
	NOTIFICATION = 3
	KEEPALIVE = 4

class BGPMessageParser:
	def __init__(self):
		pass

	@staticmethod
	def from_bytes(bbuff):
		return BGPMessage.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		pos = buff.tell()
		marker = buff.read(16)
		msg_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		msg_type = BGPMessageType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		buff.seek(pos, 0)
		msg = bgptype2class[msg_type].from_bytes(buff.read(msg_length))

		return msg

	@staticmethod
	async def from_streamreader(reader):
		t_marker = await readexactly_or_exc(reader, 16, timeout = self.timeout)
		t_length = await readexactly_or_exc(reader, 2, timeout = self.timeout)
		t_type = await readexactly_or_exc(reader, 1, timeout = self.timeout)
		msg_type = BGPMessageType(int.from_bytes(t_type, byteorder = 'big', signed = False))
		msg_length = int.from_bytes(t_length, byteorder = 'big', signed = False)
		rlen = msg_length - 19
		t_data = await readexactly_or_exc(reader, rlen, timeout = self.timeout)
		return bgptype2class[msg_type].from_bytes(t_marker + t_length + t_type + t_data)



# https://tools.ietf.org/html/rfc3392
# TODO: look for the actual types supported, the rfc doesnt say anything about that!
class BGPOptParameter:
	def __init__(self):
		self.type = None
		self.length = None
		self.value = None

	@staticmethod
	def from_bytes(bbuff):
		return BGPOptParameter.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		opt = BGPOptParameter()
		opt.type = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		opt.length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		opt.value = buff.read(opt.length)
		return opt

	def to_bytes(self):
		data = self.type.value.to_bytes(1, byteorder = 'big', signed = False)
		self.length = len(self.value)
		data += self.length.to_bytes(1, byteorder = 'big', signed = False)
		data += self.value
		return data

class BGPOpen:
	def __init__(self):
		self.marker = b'\xFF'*16
		self.msg_length = None
		self.msg_type = BGPMessageType.OPEN
		self.version = None
		self.my_AS = None
		self.hold_time = None
		self.bgp_identifier = None
		self.optional_parameters_length = None
		self.optional_parameters = []

	@staticmethod
	def from_bytes(bbuff):
		return BGPOpen.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		bo = BGPOpen()
		bo.marker = buff.read(16)
		bo.msg_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		bo.msg_type = BGPMessageType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		
		bo.version = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		bo.my_AS = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		bo.hold_time = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		bo.bgp_identifier = int.from_bytes(buff.read(4), byteorder = 'big', signed = False)
		bo.optional_parameters_length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		if bo.optional_parameters_length > 0:
			rlen = bo.optional_parameters_length
			while rlen > 0:
				opt_param = BGPOptParameter.from_buffer(buff)
				bo.optional_parameters.append(opt_param)
				rlen -= opt_param.length

		return bo

	def to_bytes(self):
		data = self.version.to_bytes(1, byteorder = 'big', signed = False)
		data += self.my_AS.to_bytes(2, byteorder = 'big', signed = False)
		data += self.hold_time.to_bytes(2, byteorder = 'big', signed = False)
		data += self.bgp_identifier.to_bytes(4, byteorder = 'big', signed = False)

		opt_data = b''
		for opt in self.optional_parameters:
			opt_data += opt.to_bytes()

		self.optional_parameters_length = len(opt_data)
		data += self.optional_parameters_length.to_bytes(1, byteorder = 'big', signed = False)
		data += opt_data


		self.msg_length = len(data)
		
		hdr  = self.marker
		hdr += self.msg_length.to_bytes(2, byteorder = 'big', signed = False)
		hdr += self.msg_type.value.to_bytes(1, byteorder = 'big', signed = False)
		return hdr + data

class WithdrawnRoute:
	def __init__(self):
		self.length = None
		self.prefix = None


	@staticmethod
	def from_bytes(bbuff):
		return WithdrawnRoute.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		wr = WithdrawnRoute()
		wr.length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		wr.prefix = buff.read(wr.length)
		return wr

	def to_bytes(self):
		return self.length.to_bytes(1, byteorder = 'big', signed = False) + self.prefix

class PathAttributeTypeFlag(enum.IntFlag):
	OPTIONAL = 128
	TRANSITIVE = 64
	PARTIAL = 32
	EXTENDED_LENGTH = 16
	UNUSED_3 = 8
	UNUSED_2 = 4
	UNUSED_1 = 2
	UNUSED_0 = 1

class PathAttributeTypeCode(enum.Enum):
	ORIGIN = 1
	AS_PATH = 2
	NEXT_HOP = 3
	MULTI_EXIT_DISC = 4
	LOCAL_PREF = 5
	ATOMIC_AGGREGATE = 6
	AGGREGATOR = 7
	COMMUNITIES = 8
	ORIGINATOR_ID = 9
	CLUSTER_LIST = 10


class PathAttributeType:
	def __init__(self):
		self.flags = None
		self.code = None

	@staticmethod
	def from_bytes(bbuff):
		return PathAttributeType.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		pat = PathAttributeType()
		pat.flags = PathAttributeTypeFlag(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		pat.code = PathAttributeTypeCode(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		return pat

	def to_bytes(self):
		return self.flags.value.to_bytes(1, byteorder = 'big', signed = False) + self.code.value.to_bytes(1, byteorder = 'big', signed = False)

class PathAttribute:
	def __init__(self):
		self.type = None
		self.length = None
		self.value = None

	@staticmethod
	def from_bytes(bbuff):
		return PathAttribute.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		pa = PathAttribute()
		pa.type = PathAttributeType.from_buffer(buff)
		if pa.type.flags & PathAttributeTypeFlag.EXTENDED_LENGTH:
			pa.length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		else:
			pa.length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		pa.value = pa.parse_value(buff.read(pa.length))
		return pa

	#override this method for implementing different types!
	def parse_value(self, bbuff):
		return bbuff

	#override this method for implementing different types!
	def serialize_value(self):
		return self.value

	def to_bytes(self):
		data = self.type.to_bytes()
		if self.type.flags & PathAttributeTypeFlag.EXTENDED_LENGTH:
			data += self.length.to_bytes(2, byteorder = 'big', signed = False)
		else:
			data += self.length.to_bytes(1, byteorder = 'big', signed = False)

		data += self.serialize_value()
		return data

class BGPNetworkLayerReachabilityInfo:
	def __init__(self):
		self.length = None
		self.prefix = None

	@staticmethod
	def from_bytes(bbuff):
		return BGPNetworkLayerReachabilityInfo.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		nlri = BGPNetworkLayerReachabilityInfo()
		nlri.length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		nlri.prefix = buff.read(nlri.length)
		return nlri

	def to_bytes(self):
		return self.length.to_bytes(1, byteorder = 'big', signed = False) + self.prefix

class BGPUpdate:
	def __init__(self):
		self.marker = b'\xFF'*16
		self.msg_length = None
		self.msg_type = BGPMessageType.UPDATE
		self.withdrawn_routes_length = None
		self.withdrawn_routes = []
		self.total_path_attribute_length = None
		self.path_attributes = []
		self.network_layer_reachability_information = None

	@staticmethod
	def from_bytes(bbuff):
		return BGPUpdate.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		bu = BGPUpdate()
		bu.marker = buff.read(16)
		bu.msg_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		bu.msg_type = BGPMessageType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))

		bu.withdrawn_routes_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		if bu.withdrawn_routes_length > 0:
			rlen = bu.withdrawn_routes_length
			while rlen > 0:
				wr = WithdrawnRoute.from_buffer(buff)
				bu.withdrawn_routes.append(wr)
				rlen -= wr.length

		bu.total_path_attribute_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		if bu.total_path_attribute_length > 0:
			rlen = bu.total_path_attribute_length
			while rlen > 0:
				pa = PathAttribute.from_buffer(buff)
				bu.path_attributes.append(pa)

		bu.network_layer_reachability_information = BGPNetworkLayerReachabilityInfo.from_buffer(buff)

		return bu

	def to_bytes(self):
		wr_data = b''
		for wr in self.withdrawn_routes:
			wr_data += wr.to_bytes()

		self.withdrawn_routes_length = len(wr_data)
		wr_len = self.withdrawn_routes_length.to_bytes(2, byteorder = 'big', signed = False)
		data = wr_len + wr_data

		pa_data = b''
		for pa in self.path_attributes:
			pa_data += pa.to_bytes()

		self.total_path_attribute_length = len(pa_data)
		tpal = self.total_path_attribute_length.to_bytes(2, byteorder = 'big', signed = False)
		data += tpal
		data += pa_data
		data += self.network_layer_reachability_information.to_bytes()
		self.msg_length = len(data)
		
		hdr  = self.marker
		hdr += self.msg_length.to_bytes(2, byteorder = 'big', signed = False)
		hdr += self.msg_type.value.to_bytes(1, byteorder = 'big', signed = False)
		return hdr + data

class BGPKeepalive:
	def __init__(self):
		self.marker = b'\xFF'*16
		self.msg_length = 19
		self.msg_type = BGPMessageType.KEEPALIVE

	@staticmethod
	def from_bytes(bbuff):
		return BGPKeepalive.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		bk = BGPKeepalive()
		bk.marker = buff.read(16)
		bk.msg_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		bk.msg_type = BGPMessageType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		return bk

	def to_bytes(self):
		t  = self.marker
		t += self.msg_length.to_bytes(2, byteorder = 'big', signed = False)
		t += self.msg_type.value.to_bytes(1, byteorder = 'big', signed = False)
		return t

class BGPError(enum.Enum):
	MESSAGE_HEADER = 1
	OPEN_MESSAGE = 2
	UPDATE_MESSAGE = 3
	HOLD_TIMER_EXPIRED = 4
	FINITE_STATE_MACHINE = 5
	CEASE = 6

class BGPErrorMessageHeader(enum.Enum):
	CONNECTION_NOT_SYNCHRONIZED = 1
	BAD_MESSAGE_LENGTH = 2
	BAD_MESSAGE_TYPE = 3

class BGPErrorOpenMessage(enum.Enum):  
	UNSUPPORTED_VERSION_NUMBER = 1
	BAD_PEER_AS = 2
	BAD_BGP_IDENTIFIER = 3
	UNSUPPORTED_OPTIONAL_PARAMETER = 4
	DEPRECATED_ERROR_CODE = 5
	UNACCEPTABLE_HOLD_TIME = 6

class BGPErrorUpdateMessage(enum.Enum):
	MALFORMED_ATTR_LIST = 1
	UNRECOGNIZED_WELL_KNOWN_ATTR = 2
	MISSING_WELL_KNOWN_ATTRIBUTE = 3
	ATTRIBUTE_FLAGS_ERROR = 4
	ATTRIBUTE_LENGTH_ERROR = 5
	INVALID_ORIGIN_ATTRIBUTE = 6
	DEPRECATED = 7
	INVALID_NEXT_HOP_ATTR = 8
	OPTIONAL_ATTRIBUTE_ERROR = 9
	INVALID_NETWORK_FILED = 10
	MALFORMED_AS_PATH = 11

class BGPNotification:
	def __init__(self):
		self.marker = b'\xFF'*16
		self.msg_length = None
		self.msg_type = BGPMessageType.NOTIFICATION
		self.error_code = None
		self.error_subcode = None
		self.data = None

	@staticmethod
	def from_bytes(bbuff):
		return BGPNotification.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		nf = BGPNotification()
		nf.marker = buff.read(16)
		nf.msg_length = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		nf.msg_type = BGPMessageType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		nf.error_code = BGPError(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		if nf.error_code == BGPError.MESSAGE_HEADER:
			nf.error_subcode = BGPErrorMessageHeader(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		elif nf.error_code == BGPError.OPEN_MESSAGE:
			nf.error_subcode = BGPErrorOpenMessage(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		elif nf.error_code == BGPError.UPDATE_MESSAGE:
			nf.error_subcode = BGPErrorUpdateMessage(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		else:
			nf.error_subcode = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)

		nf.data = buff.read(self.msg_length - 21)
		return nf

	def to_bytes(self):
		data = error_code.value.to_bytes(1, byteorder = 'big', signed = False)
		if isinstance(self.error_subcode, enum.Enum):
			data += self.error_subcode.value.to_bytes(1, byteorder = 'big', signed = False)
		else:
			data += self.error_subcode.to_bytes(1, byteorder = 'big', signed = False)
		data += self.data

		self.msg_length = len(data)
		
		hdr  = self.marker
		hdr += self.msg_length.to_bytes(2, byteorder = 'big', signed = False)
		hdr += self.msg_type.value.to_bytes(1, byteorder = 'big', signed = False)
		return hdr + data


bgptype2class = {
	BGPMessageType.OPEN : BGPOpen,
	BGPMessageType.UPDATE : BGPUpdate,
	BGPMessageType.NOTIFICATION : BGPNotification,
	BGPMessageType.KEEPALIVE : BGPKeepalive,
}


# https://tools.ietf.org/html/rfc7252
# We got packets on packets on packets on packets

import io
import enum

from responder3.core.asyncio_helpers import *

class CoAPCode(enum.Enum):
	pass

class CoAPType(enum.Enum):
	CON = 0
	NON = 1
	ACK = 2
	RST = 3

class CoAPOptionNumber(enum.Enum):
	RESERVED_0 = 0
	IF_MATCH = 1
	URI_HOST = 3 
	ETAG = 4
	IF_NONE_MATCH = 5
	URI_PORT = 7 
	LOCATION_PATH = 8 
	URI_PATH = 11 
	CONTENT_FORMAT = 12 
	MAX_AGE = 14 
	URI_QUERY = 15 
	ACCEPT = 17 
	LOCATION_QUERY = 20 
	PROXY_URI = 35
	PROXY_SCHEME = 39
	SIZE1 = 60
	RESERVED_128 = 128
	RESERVED_132 = 132
	RESERVED_136 = 136
	RESERVED_140 = 140

class CoAPOption:
	def __init__(self):
		self.delta = None
		self.length = None
		self.value = None

	@staticmethod
	def from_bytes(bbuff, option_sum = 0):
		return CoAPOption.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff, option_sum = 0):
		t = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		delta = t >> 4
		length = t & 0b00001111
		if delta > 12:
			if delta == 13:
				delta = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
			elif delta == 14:
				delta = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
			else:
				raise Exception('Packet parse error! Opt delta')

		delta += option_sum #https://tools.ietf.org/html/rfc7252#section-3
		
		if length > 12:
			if length == 13:
				length = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
			elif length == 14:
				length = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
			else:
				raise Exception('Packet parse error! Opt length')

		value = buff.read(length)
		#
		opt = a
		return opt






class CoAPPacket:
	def __init__(self):
		self.version = None
		self.type = None
		self.token_length = None
		self.code = None
		self.message_id = None
		self.token = None
		self.options = []
		self.payload = None

	@staticmethod
	def from_bytes(bbuff):
		return CoAPPacket.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		packet = CoAPPacket()
		t = int.from_bytes(buff.read(1),byteorder = 'big', signed = False)
		packet.version = t >> 6
		packet.type = (t >> 4) & 0b00000011
		packet.token_length = t & 0b00000011
		packet.code = CoAPCode(int.from_bytes(buff.read(1),byteorder = 'big', signed = False))
		packet.message_id = int.from_bytes(buff.read(2),byteorder = 'big', signed = False)
		packet.token = int.from_bytes(buff.read(packet.token_length),byteorder = 'big', signed = False)

		packet.options = []
		packet.payload = None

		
	def to_bytes(self):




class DictAuth:
	def __init__(self):
		self.credentials = None

	def setup(self, d):
		self.credentials = d['creds']

	def setup_defaults(self):
		self.credentials = {}

	def test_domain(self, domain):
		return domain in self.credentials

	def test_username(self, domain, username):
		res = self.test_domain(domain)
		if res == True:
			return username in self.credentials[domain]
		return False

	def test_password(self, plaintext_cred):
		password = self.get_password(plaintext_cred.domain, plaintext_cred.username)
		if password == False:
			return False
		if password == plaintext_cred.password:
			return True
		return False

	def get_password(self, domain, username):
		res = self.test_username(domain, username)
		if res == True:
			return self.credentials[domain][username]
		return False

	def yield_credentials(self):
		for domain in self.credentials:
			for username in self.credentials[domain]:
				yield PlaintextCredential(domain, username, password)



import re
import socket
import logging
import asyncio
import traceback
import ipaddress
import datetime
import collections

from responder3.core.interfaces.NetworkInterfaces import interfaces
from responder3.core.commons import PoisonerMode
from responder3.protocols.DNS import * 
from responder3.core.udpwrapper import UDPClient
from responder3.core.servertemplate import ResponderServer, ResponderServerSession, ResponderServerGlobalSession


class DNSGlobalSession():
	def __init__(self, listener_socket_config, settings, log_queue):
		ResponderServerGlobalSession.__init__(self, log_queue, self.__class__.__name__)
		self.listener_socket_config = listener_socket_config
		self.settings = settings
		self.spooftable = collections.OrderedDict()
		self.poisonermode = PoisonerMode.ANALYSE
		self.passthru = True
		self.passthru_iface = None
		self.passthru_iface_idx = None
		self.passthru_proto = None
		self.passthru_ip = None

		self.parse_settings()

	def parse_settings(self):
		self.passthru = False
		if 'passthru' in self.settings and self.settings['passthru']:
			self.passthru = True
			marker = self.settings['passthru']['dnsserver'].rfind(':')
			if marker != -1:
				self.passthru_server = ipaddress.ip_address(self.settings['passthru']['dnsserver'][:marker])
				self.passthru_port = int(self.settings['passthru']['dnsserver'][marker+1:])
			
			else:
				self.passthru_server = ipaddress.ip_address(self.settings['passthru']['dnsserver'])
				self.passthru_port   = 53
			
			self.passthru_iface  = self.settings['passthru']['bind_iface'] if 'bind_iface' in self.settings['passthru'] else self.listener_socket_config.bind_iface
			self.passthru_iface_idx = self.listener_socket_config.bind_iface_idx
			self.passthru_proto  = self.settings['passthru']['bind_protocol'] if 'bind_protocol' in self.settings['passthru'] else self.listener_socket_config.bind_protocol
			self.passthru_ip     = ipaddress.ip_address(self.settings['passthru']['bind_addr']) if 'bind_addr' in self.settings['passthru'] else None

			if self.passthru_ip is None and self.passthru_iface != self.listener_socket_config.bind_iface:
				self.passthru_ip = interfaces.get_ip(self.passthru_iface, self.passthru_server.version)[0]
			else:
				self.passthru_ip = self.listener_socket_config.bind_addr

		if self.settings is None:
			self.log('No settings defined, adjusting to Analysis functionality!')

		else:
			# parse the poisoner mode
			if isinstance(self.settings['mode'], str):
				self.poisonermode = PoisonerMode[self.settings['mode'].upper()]

			# compiling re strings to actual re objects and converting IP strings to IP objects
			if self.poisonermode == PoisonerMode.SPOOF:
				for entry in self.settings['spooftable']:
					for regx in entry:
						self.spooftable[re.compile(regx)] = ipaddress.ip_address(entry[regx])

class DNSSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)


class DNS(ResponderServer):
	def init(self):
		self.parser = DNSPacket

	async def parse_message(self):
		msg = await asyncio.wait_for(self.parser.from_streamreader(self.creader, self.globalsession.listener_socket_config.bind_protocol), timeout = 1)
		return msg

	async def send_data(self, data, addr = None):
		if self.listener_socket_config.bind_protocol == socket.SOCK_DGRAM:
			self.cwriter.write(data)
			#await asyncio.wait_for(self.cwriter.write(data, addr), timeout=1)
		else:
			self.cwriter.write(data)
			await self.cwriter.drain()
		return

	async def poll_dnsserver(self, msg):
		await self.log('Polling passthru server at %s!' % (str(self.globalsession.passthru_server),))
		if self.globalsession.passthru_proto == socket.SOCK_DGRAM:
			if self.globalsession.passthru_ip.version == 4:
				sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
				sock.setblocking(False)
				sock.setsockopt(socket.SOL_SOCKET, 25, self.globalsession.passthru_iface.encode())
				sock.bind((str(self.globalsession.passthru_ip), 0))
			
			else:
				sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
				sock.setblocking(False)
				sock.setsockopt(socket.SOL_SOCKET, 25, self.globalsession.passthru_iface.encode())
				sock.bind((str(self.globalsession.passthru_ip), 0, self.globalsession.passthru_iface_idx ))

			client = UDPClient((str(self.globalsession.passthru_server), self.globalsession.passthru_port), sock = sock)
			reader, writer = await asyncio.wait_for(client.run(msg.to_bytes()), timeout = 1)

		else:
			if self.globalsession.passthru_ip.version == 4:
				sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP)
				sock.setsockopt(socket.SOL_SOCKET, 25, self.globalsession.passthru_iface.encode())
				sock.bind((str(self.globalsession.passthru_ip), 0))
				sock.connect((str(self.globalsession.passthru_server), self.globalsession.passthru_port))

			else:
				sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_TCP)
				sock.setblocking(False)
				sock.setsockopt(socket.SOL_SOCKET, 25, self.globalsession.passthru_iface.encode())
				sock.bind((str(self.globalsession.passthru_ip), 0, self.globalsession.passthru_iface_idx ))
				sock.connect((str(self.globalsession.passthru_server), self.globalsession.passthru_port))
			
			reader, writer = await asyncio.wait_for(asyncio.open_connection(sock = sock), timeout = 1)
		
			writer.write(msg.to_bytes())
			await writer.drain()
		
		passthru_packet = await asyncio.wait_for(DNSPacket.from_streamreader(reader, self.globalsession.passthru_proto), timeout = 1)
		await self.log('Passthru packet recieved!')
		return passthru_packet

	async def run(self):
		try:
			msg = await asyncio.wait_for(self.parse_message(), timeout=1)
			if self.globalsession.poisonermode == PoisonerMode.ANALYSE:
				for q in msg.Questions:
					await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, request_type = q.QTYPE.name)

			else:
				answers = []
				for q in msg.Questions:
					for spoof_regx in self.globalsession.spooftable:
						spoof_ip = self.globalsession.spooftable[spoof_regx]
						if spoof_regx.match(q.QNAME.name):
							res = None
							await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, poisonName = str(spoof_regx), poisonIP = spoof_ip, request_type = q.QTYPE.name)
							if spoof_ip.version == 4 and q.QTYPE == DNSType.A:
								res = DNSAResource.construct(q.QNAME.name, spoof_ip)
							elif spoof_ip.version == 6 and q.QTYPE == DNSType.AAAA:
								res = DNSAAAAResource.construct(q.QNAME.name, spoof_ip)
							
							if res:
								answers.append(res)
							break

						elif self.globalsession.passthru:
							await self.logger.poisonresult(PoisonerMode.ANALYSE, requestName = q.QNAME.name, request_type = q.QTYPE.name)
							#if ANY of the query names requested doesnt match our spoof table, then we ask an actual DNS server
							#this completely overrides any match from the spooftable!
							passthru_ans = await asyncio.wait_for(self.poll_dnsserver(msg), timeout=1)
							#modify response here!
							passthru_ans_modified = passthru_ans
							#print(passthru_ans_modified)
							await asyncio.wait_for(self.send_data(passthru_ans_modified.to_bytes()), timeout=1)
							return
						
						else:
							await self.logger.poisonresult(PoisonerMode.ANALYSE, requestName = q.QNAME.name, request_type = q.QTYPE.name)

				if len(answers) == 0 :
					#DNS error response should be here!
					#raise Exception('DNS error response should be here!')
					return

				response = DNSPacket.construct(TID = msg.TransactionID, 
												 response = DNSResponse.RESPONSE, 
												 answers = answers,
												 questions = msg.Questions,
												 proto = self.listener_socket_config.bind_protocol)
			
				await asyncio.wait_for(self.send_data(response.to_bytes()), timeout=1)

		except Exception as e:
			raise e


# https://www.rfc-editor.org/rfc/rfc3261.txt

import io
import enum
import asyncio
import traceback

from responder3.core.logging.log_objects import Credential
from responder3.core.commons import read_element
from responder3.core.asyncio_helpers import *

class SIPSessionStatus:
	UNAUTHENTICATED = enum.auto()
	AUTHENTICATED = enum.auto()
	AUTHFAILED  = enum.auto()

class VIAHeader:
	def __init__(self):
		self.transport = None
		self.sent_by_addr = None
		self.received = None
		self.rport = None
		self.branch = None

class AUTHORIZATIONHeader:
	def __init__(self):
		#Digest username="asdfashfgha", realm="sip.cybercity.dk", nonce="1701af566be182070084c6f740706bb", 
		#uri="sip:192.168.42.211:5066", response="713996be6fe3557c4400f28fb10977b8", algorithm=MD5, opaque="1701a1351f70795"
		self.auth_type = None
		self.username = None
		self.realm = None
		self.nonce = None
		self.uri = None
		self.response = None
		self.algorithm = None
		self.opaque = None

		self.unknown_tags = {}

	@staticmethod
	def from_bytes(bbuff):
		ah = AUTHORIZATIONHeader()
		m = bbuff.find(' ')
		if m == -1:
			raise Exception('Wrong Authroization header format!')
		ah.auth_type = bbuff[:m].strip()
		tags = bbuff[m+1:].split(',')
		for raw_tag in tags:
			tag = rawtag.strip()
			m = tag.find('=')
			if m == -1:
				raise Exception('Wrong Authroization header format!')
			tag_key = tag[:m]
			tag_value = tag[m+1:].strip()
			if tag_value[0] == '"':
				tag_value = tag_value[1:-1]

			if tag_key.lower() == 'username':
				self.username = tag_value
			elif tag_key.lower() == 'realm':
				self.realm = tag_value
			elif tag_key.lower() == 'nonce':
				self.nonce = tag_value
			elif tag_key.lower() == 'uri':
				self.uri = tag_value
			elif tag_key.lower() == 'response':
				self.response = tag_value
			elif tag_key.lower() == 'algorithm':

				self.algorithm = tag_value
			elif tag_key.lower() == 'opaque':
				self.opaque = tag_value
			else:
				self.unknown_tags[tag_key] = tag_value


	def get_sip_hash(self, method = 'REGISTER'):
		#$sip$*192.168.100.100*192.168.100.121*username*asterisk*REGISTER*sip*192.168.100.121**2b01df0b****MD5*ad0520061ca07c120d7e8ce696a6df2d 
		return '$sip$*%s*%s*%s*%s*%s*%s*%s*%s*%s*%s*%s*%s*%s*%s' % (self.uri, '', self.username, self.realm, method, self.uri_prefix, '', '', self.nonce, '','', self.algorithm, self.response)

		#$sip$*[URI_SERVER]*[URI_CLIENT]*[USERNAME]*[REALM]*[METHOD]*[URI_PREFIX]*[URI_RESOURCE]*[URI_SUFFIX]*[NONCE_SERVER]*[NONCE_CLIENT]*[NONCE_COUNT]*[QOP]*[DIRECTIVE]*[MD5]
		#$sip$*sipConfusedip.aaa.it**33333*sip.aaa.it*REGISTER*sip*sip.aaa.it**1234****MD5*7468b74b7257b05671242ad0a8b0eb16'


class Request:
	def __init__(self):
		self.method = None
		self.request_uri = None
		self.sip_version = None

		self.headers = {}
		self.data = None

		#special headers for making this module lighter (otherwise we'd implement the headers as objects)
		self.spec_headers = {}

	@staticmethod
	async def from_streamreader(reader, timeout = 10):
		req = Request()
		line = await readline_or_exc(reader, timeout)
		req.method, req.request_uri, req.sip_version = line.decode().strip().split(' ')

		hdr_blob = await readuntil_or_exc(reader, b'\r\n\r\n', timeout)
		hdr_blob = hdr_blob.decode()
		hdr_lines = hdr_blob.split('\r\n')
		
		prev_key = None
		for line in hdr_lines:
			if line == '':
				continue
			m = line.find(':')
			if m == -1:
				self.headers[prev_key] += line.strip()
			else:
				key = line[:m].strip()
				prev_key = key
				req.headers[key] = line[m+1:].strip()
				req.spec_headers[key.lower()] = line[m+1:].strip()

		if 'content-length' in req.spec_headers:
			req.data = await readexactly_or_exc(reader, int(req.spec_headers['content-length']), timeout)

		return req

	def __str__(self):
		t = '============ SIP Request ================\r\n'
		t += ' '.join([self.method, self.request_uri, self.sip_version])
		t += '\r\n'
		for key in self.headers:
			t += '%s: %s\r\n' % (key, self.headers[key])
		if self.data:
			t += self.data.hex()
		return t


class Response:
	def __init__(self):
		self.sip_version = None
		self.status_code = None
		self.reason = None

		self.headers = {}
		self.data = None

	def to_bytes(self):
		status_line = ' '.join([self.sip_version, self.status_code, self.reason])
		t = status_line + '\r\n'
		for key in self.headers:
			t += '%s: %s\r\n' % (key, self.headers[key])
		t+= '\r\n'
		t = t.encode()
		if self.data:
			t += data
		return t

	@staticmethod
	async def from_streamreader(reader, timeout = 10):
		resp = Response()
		line = await readline_or_exc(reader, timeout)
		resp.sip_version, resp.status_code, resp.reason = line.decode().strip().split(' ')

class SIP401Response(Response):
	def __init__(self):
		Response.__init__(self)
		self.sip_version = 'SIP/2.0'
		self.status_code = '401'
		self.reason = 'Unauthorized'

	@staticmethod
	def from_request(req, auth_data):
		resp = SIP401Response()
		for key in ['Call-ID', 'CSeq', 'From', 'To', 'Via']:
			resp.headers[key] = req.headers[key]
		resp.headers['WWW-Authenticate'] = auth_data
		resp.headers['Content-Length'] = 0
		return resp

class SIP403Response(Response):
	def __init__(self):
		Response.__init__(self)
		self.sip_version = 'SIP/2.0'
		self.status_code = '403'
		self.reason = 'Forbidden'

	@staticmethod
	def from_request(req, auth_data):
		resp = SIP403Response()
		resp.headers['Content-Length'] = 0
		return resp



import logging
import asyncio
from urllib.parse import urlparse
import uuid
import random
from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import *
from responder3.core.commons import *
from responder3.protocols.SMB.SMBTransport import SMBTransport
from responder3.protocols.SMB.SMB import * 
from responder3.protocols.SMB.SMB2 import * 
from responder3.protocols.SMB.ntstatus import * 
from responder3.protocols.authentication.GSSAPI import * 
from responder3.protocols.authentication.common import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession

class SMB2ServerState(enum.Enum):
	UNAUTHENTICATED = enum.auto()
	AUTHENTICATED = enum.auto()

class SMBSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)

		self.parser = SMBTransport
		self.SMBprotocol  = None
		#self.SMBdialect   = 'SMB 2.002'
		self.SMBdialect   = ['SMB 2.002', 'NT LM 0.12', ]
		self.commondialect = None

		self.current_state = SMB2ServerState.UNAUTHENTICATED
		self.gssapihandler = GSSAPIAuthHandler()
		self.serverUUID = uuid.UUID(bytes=os.urandom(16))
		self.SMBSessionID = os.urandom(8)
		self.SMBMessageCnt = 0
		#self.authAPI      = None


	def __repr__(self):
		t  = '== SMBSession ==\r\n'
		t += 'SMBprotocol:      %s\r\n' % repr(self.SMBprotocol)
		t += 'SMBdialect: %s\r\n' % repr(self.SMBdialect)
		t += 'commondialect: %s\r\n' % repr(self.commondialect)
		t += 'current_state: %s\r\n' % repr(self.current_state)
		t += 'gssapihandler:       %s\r\n' % repr(self.gssapihandler)
		t += 'serverUUID: %s\r\n' % repr(self.serverUUID)
		t += 'SMBSessionID:     %s\r\n' % repr(self.SMBSessionID)
		t += 'SMBMessageCnt:     %s\r\n' % repr(self.SMBMessageCnt)
		return t


class SMB(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		return

	def parse_settings(self):
		#TODO
		self.set_default_settings()



	async def parse_message(self, timeout = 10):
		try:
			smbtransport = await asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout = timeout)
			return smbtransport.smbmessage
		except asyncio.TimeoutError:
			await self.log('Timeout!', logging.DEBUG)

	async def send_data(self, smbmessage):
		data = self.session.parser.construct(smbmessage)
		self.cwriter.write(data.to_bytes())
		await self.cwriter.drain()

	async def send_authfailed(self, smbmessage):
		return

	@r3trafficlogexception
	async def run(self):
		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0].smbmessage

			if self.session.current_state == SMB2ServerState.UNAUTHENTICATED:
				#this could be SMB/NegotiateProtocol or SMB2/NegotiateProtocol or SMB2/SessionSetup
				if msg.type == 1:
					if msg.header.Command == SMBCommand.SMB_COM_NEGOTIATE:
						#the first message could be SMBv1/NegotiateProtocol to identiy smbv2 capabilities
						await self.logger.debug([dialect.DialectString for dialect in msg.command.Dialects])
						#selecting the dialect which is common to both server and client in order of server dialect list prefereances
						clinet_dialects = set([dialect.DialectString for dialect in msg.command.Dialects])
						server_dialects = set(self.session.SMBdialect)
						common_dialects = server_dialects.intersection(clinet_dialects)
						if common_dialects is None:
							await self.logger.info('No matching dialects between client and server, terminating connection!')
							return
						preferred_dialect = None
						for dialect in self.session.SMBdialect:
							for cd in common_dialects:
								if cd == dialect:
									preferred_dialect = dialect
									break
							else:
								continue
							break
						#now, for smbv1 we need to get the index of the selected dialect... 
						preferred_dialect_idx = 0
						for dialect in [dialect.DialectString for dialect in msg.command.Dialects]:
							if dialect == preferred_dialect:
								#preferred_dialect_idx += 1
								break
							preferred_dialect_idx += 1

						await self.logger.info(preferred_dialect_idx)

						self.session.commondialect = preferred_dialect

						if self.session.commondialect == 'SMB 2.002':
							await self.logger.debug('client is capable of smbv2, using SMBv2')
							status, data, t = self.session.gssapihandler.do_auth()
							resp = SMB2Message()
							resp.header = SMB2Header_ASYNC.construct(SMB2Command.NEGOTIATE, SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR, self.session.SMBMessageCnt)
							resp.command = NEGOTIATE_REPLY.construct(data, NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED,
								NegotiateDialects.SMB202, self.session.serverUUID, NegotiateCapabilities.SMB2_GLOBAL_CAP_DFS|NegotiateCapabilities.SMB2_GLOBAL_CAP_LEASING|NegotiateCapabilities.SMB2_GLOBAL_CAP_LARGE_MTU)

							# MessageCnt should not be incremented here
							a = await asyncio.wait_for(self.send_data(resp), timeout=1)
							
						else:
							await self.logger.debug('using SMBv1')
							status, data, t = self.session.gssapihandler.do_auth(smbv1 = True)
							resp = SMBMessage()
							resp.header = SMBHeader.construct(SMBCommand.SMB_COM_NEGOTIATE, 
																NTStatus.STATUS_SUCCESS, 
																	
																SMBHeaderFlagsEnum.SMB_FLAGS_REPLY|
																SMBHeaderFlagsEnum.SMB_FLAGS_CASE_INSENSITIVE, 
																SMBHeaderFlags2Enum.SMB_FLAGS2_UNICODE|
																SMBHeaderFlags2Enum.SMB_FLAGS2_EXTENDED_SECURITY|
																SMBHeaderFlags2Enum.SMB_FLAGS2_NT_STATUS|
																SMBHeaderFlags2Enum.SMB_FLAGS2_LONG_NAMES,
																mid = msg.header.MID,
																pidhigh = msg.header.PIDHigh,
																pidlow = msg.header.PIDLow,
															)
																	

							resp.command = SMB_COM_NEGOTIATE_REPLY.construct(preferred_dialect_idx, 
															SMBSecurityMode.NEGOTIATE_ENCRYPT_PASSWORDS|SMBSecurityMode.NEGOTIATE_USER_SECURITY, 
															b'\x00'*4, 
															SMBCapabilities.CAP_UNICODE|SMBCapabilities.CAP_NT_SMBS|SMBCapabilities.CAP_LARGE_FILES|SMBCapabilities.CAP_NT_EXTENDED_SECURITY, 
															self.session.serverUUID, data,
														)

							a = await asyncio.wait_for(self.send_data(resp), timeout=1)
							continue
					else:
						if msg.header.Command == SMBCommand.SMB_COM_SESSION_SETUP_ANDX:
							status, data, cred = self.session.gssapihandler.do_auth(msg.command.SecurityBlob)

							if status in [AuthResult.FAIL, AuthResult.OK]:
								await self.logger.credential(cred.to_credential())

								
							if status == AuthResult.CONTINUE:
								resp.header = SMBHeader.construct(SMBCommand.SMB_COM_SESSION_SETUP_ANDX, 
																NTStatus.STATUS_MORE_PROCESSING_REQUIRED, 
																	
																SMBHeaderFlagsEnum.SMB_FLAGS_REPLY|
																SMBHeaderFlagsEnum.SMB_FLAGS_CASE_INSENSITIVE, 
																SMBHeaderFlags2Enum.SMB_FLAGS2_UNICODE|
																SMBHeaderFlags2Enum.SMB_FLAGS2_EXTENDED_SECURITY|
																SMBHeaderFlags2Enum.SMB_FLAGS2_NT_STATUS|
																SMBHeaderFlags2Enum.SMB_FLAGS2_LONG_NAMES,
																uid = random.randint(1,5000),
																mid = msg.header.MID,
																pidhigh = msg.header.PIDHigh,
																pidlow = msg.header.PIDLow,
															)
								resp.command = SMB_COM_SESSION_SETUP_ANDX_REPLY.construct(secblob = data, 
																						  nativeos = 'Windows 2003', 
																						  nativelanman = 'blabla'
																						)
								a = await asyncio.wait_for(self.send_data(resp), timeout=1)
								continue

							elif status == AuthResult.FAIL:							
								resp.header = SMBHeader.construct(SMBCommand.SMB_COM_SESSION_SETUP_ANDX, 
																NTStatus.STATUS_ACCOUNT_DISABLED, 
																
																SMBHeaderFlagsEnum.SMB_FLAGS_REPLY|
																SMBHeaderFlagsEnum.SMB_FLAGS_CASE_INSENSITIVE, 
																SMBHeaderFlags2Enum.SMB_FLAGS2_UNICODE|
																SMBHeaderFlags2Enum.SMB_FLAGS2_EXTENDED_SECURITY|
																SMBHeaderFlags2Enum.SMB_FLAGS2_NT_STATUS|
																SMBHeaderFlags2Enum.SMB_FLAGS2_LONG_NAMES,
																uid = msg.header.UID,
																mid = msg.header.MID,
																pidhigh = msg.header.PIDHigh,
																pidlow = msg.header.PIDLow,
															)
								resp.command = SMB_COM_SESSION_SETUP_ANDX_REPLY.construct(secblob = data, 
																						  nativeos = 'Windows 2003', 
																						  nativelanman = 'blabla'
																						)
								a = await asyncio.wait_for(self.send_data(resp), timeout=1)
								return
							
							elif status == AuthResult.OK:
								self.session.current_state = SMB2ServerState.AUTHENTICATED
								resp.header = SMBHeader.construct(SMBCommand.SMB_COM_SESSION_SETUP_ANDX, 
																NTStatus.STATUS_SUCCESS, 
																	
																SMBHeaderFlagsEnum.SMB_FLAGS_REPLY|
																SMBHeaderFlagsEnum.SMB_FLAGS_CASE_INSENSITIVE, 
																SMBHeaderFlags2Enum.SMB_FLAGS2_UNICODE|
																SMBHeaderFlags2Enum.SMB_FLAGS2_EXTENDED_SECURITY|
																SMBHeaderFlags2Enum.SMB_FLAGS2_NT_STATUS|
																SMBHeaderFlags2Enum.SMB_FLAGS2_LONG_NAMES,
																uid = msg.header.UID,
																mid = msg.header.MID,
																pidhigh = msg.header.PIDHigh,
																pidlow = msg.header.PIDLow,
															)
								resp.command = SMB_COM_SESSION_SETUP_ANDX_REPLY.construct(
									secblob = data,
									nativeos = 'Windows 2003',
									nativelanman = 'blabla'
								)
								a = await asyncio.wait_for(self.send_data(resp), timeout=1)
								continue

								
						else:
							raise Exception('not implemented!')

					
				else:
					if msg.header.Command == SMB2Command.NEGOTIATE:
						status, data, t = self.session.gssapihandler.do_auth()
						resp = SMB2Message()
						resp.header = SMB2Header_ASYNC.construct(SMB2Command.NEGOTIATE, SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR, self.session.SMBMessageCnt)
						resp.command = NEGOTIATE_REPLY.construct(data, NegotiateSecurityMode.SMB2_NEGOTIATE_SIGNING_ENABLED,
							NegotiateDialects.SMB202, self.session.serverUUID, 
							NegotiateCapabilities.SMB2_GLOBAL_CAP_DFS|NegotiateCapabilities.SMB2_GLOBAL_CAP_LEASING|NegotiateCapabilities.SMB2_GLOBAL_CAP_LARGE_MTU
							)

						a = await asyncio.wait_for(self.send_data(resp), timeout=1)
					
					elif msg.header.Command == SMB2Command.SESSION_SETUP:
						status, data, cred = self.session.gssapihandler.do_auth(msg.command.Buffer)
						if cred is not None:
							await self.logger.credential(cred.to_credential())
							
						resp = SMB2Message()
						resp.header = SMB2Header_ASYNC.construct(SMB2Command.SESSION_SETUP, SMB2HeaderFlag.SMB2_FLAGS_SERVER_TO_REDIR, 1 ,status = NTStatus.STATUS_MORE_PROCESSING_REQUIRED,
							Credit = 1, CreditCharge = 1, SessionId= self.session.SMBSessionID)
						resp.command = SESSION_SETUP_REPLY.construct(data,0)

						a = await asyncio.wait_for(self.send_data(resp), timeout=1)
					else:
						raise Exception('Dunno!')

			if self.session.current_state == SMB2ServerState.AUTHENTICATED:
				raise Exception('Dunno what to do now that authentication was sucsessfull')




import json
import enum
import datetime
import time

from responder3.core.commons import *

class LogObjectType(enum.Enum):
	LOGENTRY = 0
	PROXYDATA = 1
	CONNECTION = 3
	CONNECTION_OPENED = 4
	CONNECTION_CLOSED = 5
	CREDENTIAL = 6
	POISONRESULT = 7
	EMAILENTRY = 8
	TRAFFICLOG = 9
	REMOTELOG = 10

	
class RemoteLog:
	def __init__(self, rlog):
		self.remote_ip = rlog.remote_ip
		self.remote_port = rlog.remote_port
		self.client_id = rlog.client_id
		self.log_obj = logobj2type[LogObjectType(rlog.log_obj_type)].from_dict(rlog.log_obj)
		
	def __str__(self):
		return "[%s][%s:%s] %s" % (self.client_id, self.remote_ip, self.remote_port, str(self.log_obj))

	def to_dict(self):
		t = {}
		t['remote_ip'] = self.remote_ip
		t['remote_port'] = self.remote_port
		t['client_id'] = self.client_id
		t['log_obj'] = self.log_obj.to_dict()
		return t 

	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
		
class LogEntry:
	"""
	Communications object that is used to pass log information to the LogProcessor
	"""
	def __init__(self, level, name, msg, connection = None):
		"""

		:param level: log level
		:type level: int
		:param name: name of the module emitting the message
		:type name: str
		:param msg: the message which will be logged
		:type msg: str
		"""
		self.level = level
		self.name  = name
		self.msg   = msg
		self.connection = connection

	def to_dict(self):
		t = {}
		t['level'] = self.level
		t['name'] = self.name
		t['msg'] = self.msg
		t['connection'] = self.connection.to_dict() if self.connection else None
		return t

	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)

	def __str__(self):
		t = '[%s]' % self.name
		if self.connection:
			t += '[%s]' % self.connection.get_local_print_address()
			if self.connection.remote_ip:
				t += '[%s]' % self.connection.get_remote_print_address()

		t += ' %s' % self.msg

		return t

		
	@staticmethod
	def from_dict(d):
		if 'connection' in d and d['connection']:
			return LogEntry( d['level'], d['name'], d['msg'], Connection.from_dict(d['connection']))
		else:
			return LogEntry( d['level'], d['name'], d['msg'], None)
		
		
	@staticmethod
	def from_json(data):
		return LogEntry.from_dict(json.loads(data))



class ProxyDataType(enum.Enum):
	"""
	The type of the data being logged. This is used for re-parsing the communication from the log file
	"""
	BINARY = 0
	HTTP   = 1
	SOCKS5 = 2
	SOCKS4 = 3
	FTP    = 4
	SMTP   = 5


class ProxyData:
	def __init__(self):
		"""
		Describes the intercepted communication data.
		Used to store or to read back the intercepted comms.
		"""
		self.src_addr  = None
		self.dst_addr  = None
		self.proto     = None
		self.isSSL     = None
		self.timestamp = datetime.datetime.utcnow()
		self.data_type = None
		self.data      = None

	def to_dict(self):
		"""
		Converts the object to a dictionary
		:return: dict
		"""
		t = {}
		t['src_addr'] = [str(self.src_addr[0]), int(self.src_addr[1])]
		t['dst_addr'] = [str(self.dst_addr[0]), int(self.dst_addr[1])]
		t['proto'] = self.proto.value
		t['isSSL'] = self.isSSL
		t['timestamp'] = self.timestamp
		t['data_type'] = self.data_type.value
		if self.data_type == ProxyDataType.BINARY:
			t['data'] = self.data.hex()
		else:
			raise Exception('Data type %s not implemented!' % (self.data_type))
		return t

	@staticmethod
	def from_dict(d):
		"""
		Loads the object from a dictionary
		:param d: The dictionary containing all fileds of the object
		:type d: dict
		:return: ProxyData
		"""
		pd = ProxyData()
		pd.src_addr  = (ipaddress.ip_address(d['src_addr'][0]), int(d['src_addr'][1]))
		pd.dst_addr  = (ipaddress.ip_address(d['dst_addr'][0]), int(d['dst_addr'][1]))
		pd.proto     = ServerProtocol(d['proto'])
		pd.isSSL     = bool(d['isSSL'])
		pd.timestamp = isoformat2dt(d['timestamp'])
		pd.data_type = ProxyDataType(d['data_type'])
		
		if pd.data_type == ProxyDataType.BINARY:
			pd.data = bytes.fromhex(d['data'])
		else:
			raise Exception('Data type %s not implemented!' % (pd.data_type))

		return pd

	def to_json(self):
		"""
		Used to serialize the ProxyData object
		:return: str
		"""
		return json.dumps(self.to_dict(), cls=UniversalEncoder)

	@staticmethod
	def from_json(s):
		"""
		Deserializes the ProxyData object
		:param s: JSON formatted string
		:type s: str
		:return: ProxyData
		"""
		return ProxyData.from_dict(json.loads(s))

	def __str__(self):
		if self.data_type == ProxyDataType.BINARY:
			return '[%s] [%s -> %s]\r\n%s' % (self.timestamp.isoformat(),
													'%s:%d' % self.src_addr, '%s:%d' % self.dst_addr,
													hexdump(self.data))
		else:
			raise Exception('Data type %s not implemented!' % (self.data_type))

		
class Credential:
	def __init__(self, credtype, domain = None, username = None, password = None, fullhash = None, module = None, connection = None):
		"""
		Credential object is used to log captured credential.
		This is the container for all captured credential info
		:param credtype: The type of the credential
		:type credtype: str
		:param domain: Domain info
		:param username: Username
		:type username: str
		:param password: Password
		:type password: str
		:param fullhash: The full captured credential in any format that is supported by major password crackers
		:type fullhash: str
		"""
		self.module = module
		self.connection = connection
		self.credtype = credtype
		self.domain   = domain
		self.username = username
		self.password = password
		self.fullhash     = fullhash
		self.module   = None
		self.client_addr  = None
		self.client_rdns  = None
		self.fingerprint = None

	def to_dict(self):
		"""
		Converts the object to a dict
		:return: dict
		"""
		t = {}
		t['module'] = self.module
		t['connection'] = self.connection.to_dict()
		t['credtype'] = self.credtype
		t['domain'] = self.domain
		t['username'] = self.username
		t['password'] = self.password
		t['fullhash'] = self.fullhash
		t['module'] = self.module
		t['client_addr'] = self.client_addr
		t['client_rdns'] = self.client_rdns
		t['fingerprint'] = self.fingerprint
		return t

	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
		
		
	@staticmethod
	def from_dict(d):
		cc = Credential( d['credtype'])
		cc.module = d['module']
		cc.connection = Connection.from_dict(d['connection'])
		cc.domain   = d['domain']
		cc.username = d['username']
		cc.password = d['password']
		cc.fullhash     = d['fullhash']
		cc.module   = d['module']
		cc.client_addr  = d['client_addr']
		cc.client_rdns  = d['client_rdns']
		cc.fingerprint = d['fingerprint']
		return cc
		
	@staticmethod
	def from_json(data):
		return Credential.from_dict(json.loads(data))

	def to_credential(self):
		return self

	def __str__(self):
		return '[%s][%s][%s][%s] %s' % (self.module, self.connection.get_local_print_address(), self.connection.get_remote_print_address(), 'CREDENTIAL', self.fullhash)


class PoisonResult:
	def __init__(self,connection, module = None):
		"""
		Container for messages captured or emitted by poisoner modules
		"""
		self.module = module
		self.connection = connection
		self.target = None
		self.request_name = None
		self.request_type = None
		self.poison_name = None
		self.poison_addr = None
		self.mode = None

	def to_dict(self):
		t = self.connection.to_dict()
		t['module'] = self.module
		t['connection'] = self.connection.to_dict()
		t['target'] = self.target
		t['request_name'] = self.request_name
		t['request_type'] = self.request_type
		t['poison_name'] = self.poison_name
		t['poison_addr'] = self.poison_addr
		t['mode'] = self.mode
		return t

	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)

	def __repr__(self):
		return str(self)
		
	@staticmethod
	def from_dict(d):
		cc = PoisonResult()
		cc.connection = Connection.from_dict(d['connection'])
		cc.module = d['module']
		cc.target = d['target']
		cc.request_name = d['request_name']
		cc.request_type = d['request_type']
		cc.poison_name = d['poison_name']
		cc.poison_addr = d['poison_addr']
		cc.mode = PoisonerMode(d['mode'])
		
		return cc
		
	@staticmethod
	def from_json(data):
		return PoisonResult.from_dict(json.loads(data))

	def __str__(self):
		return '%s %s %s' % (self.credtype, self.domain, self.fullhash)

	def __str__(self):
		if self.mode == PoisonerMode.ANALYSE:
			return '[%s] Recieved request from IP: %s to resolve: %s type: %s' % (self.module, self.target, self.request_name, self.request_type)
		else:
			return '[%s] Spoofing target: %s for the request: %s of type: %s which matched the expression %s. Spoof address %s' % (self.module, self.target, self.request_name, self.request_type, self.poison_name, self.poison_addr)


class EmailEntry:
	def __init__(self, connection, module = None):
		"""
		Container for emails captured
		"""
		self.module = module
		self.connection = connection
		self.fromAddress = None #string
		self.toAddress   = None #list
		self.email       = None #email object (from the email package)



class ConnectionOpened:
	def __init__(self, connection, module = None):
		self.module = module
		self.connection = connection
		
	def to_dict(self):
		t = self.connection.to_dict()
		t['module'] = self.module
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls = UniversalEncoder)
		
	@staticmethod
	def from_dict(d):
		return ConnectionOpened(Connection.from_dict(d), module= d['module'])
		
	@staticmethod
	def from_json(data):
		return ConnectionOpened.from_dict(json.loads(data))
		
	def __str__(self):
		return '[%s][%s][%s][%s]' % (self.module, self.connection.get_local_print_address(), self.connection.get_remote_print_address(), 'OPENED')

class ConnectionClosed:
	def __init__(self, connection, module = None):
		self.module = module
		self.connection = connection
		self.disconnect_time = datetime.datetime.utcnow()
		self.total_connection_time_s = (self.disconnect_time - self.connection.timestamp).total_seconds()
		
	def to_dict(self):
		t = self.connection.to_dict()
		t['total_connection_time_s'] = self.total_connection_time_s
		t['disconnect_time'] = self.disconnect_time
		t['module'] = self.module
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls = UniversalEncoder)
		
	@staticmethod
	def from_dict(d):
		cc = ConnectionClosed( Connection.from_dict(d), module= d['module'])
		cc.disconnect_time = isoformat2dt(d['disconnect_time'])
		cc.total_connection_time_s = d['total_connection_time_s']
		return cc
		
	@staticmethod
	def from_json(data):
		return ConnectionClosed.from_dict(json.loads(data))
		
	def __str__(self):
		return '[%s][%s][%s][%s] Lasted %s seconds' % (self.module, self.connection.get_local_print_address(), self.connection.get_remote_print_address(), 'CLOSED', self.total_connection_time_s)

class TrafficLog:
	def __init__(self):
		self.module = None
		self.connection = None
		self.data_recv = {}
		self.unconsumed_buffer = b''
		self.data_sent = {}

	def to_dict(self):
		t = {}
		t['module'] = self.module
		t['connection'] = self.connection.to_dict()
		t['data_recv'] = {}
		t['data_sent'] = {}
		for date in self.data_recv:
			t['data_recv'][int(date.timestamp()*10**3)] = self.data_recv[date].hex()
		for date in self.data_sent:
			t['data_sent'][int(date.timestamp()*10**3)] = self.data_sent[date].hex()
		t['unconsumed_buffer'] = self.unconsumed_buffer.hex()
		return t

	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)

	@staticmethod
	def from_dict(d):
		t = TrafficLog()
		t.module = d['module']
		t.connection = Connection.from_dict(d['connection'])
		for date in d['data_recv']:
			t.data_recv[datetime.datetime.fromtimestamp(int(date)/10**3)] = bytes.fromhex(d['data_recv'][date])
		for date in d['data_sent']:
			t.data_sent[datetime.datetime.fromtimestamp(int(date)/10**3)] = bytes.fromhex(d['data_sent'][date])
		t.unconsumed_buffer = bytes.fromhex(d['unconsumed_buffer'])
		return t

	@staticmethod
	def from_json(data):
		return TrafficLog.from_dict(json.loads(data))

	def merge(self):
		allcom = []
		for date in self.data_recv:
			allcom.append((date, 'RECV', self.data_recv[date]))
		for date in self.data_sent:
			allcom.append((date, 'SENT', self.data_sent[date]))

		return sorted(allcom, key=lambda x: x[0])



	def get_loglines(self):
		for date, direction, data in self.merge():
			yield '[%s][%s][%s][TRAFFICLOG][%s][%s] %s' % (self.module, self.connection.get_local_print_address(), self.connection.get_remote_print_address(), direction, date, data.hex())

### needs to be at the bottom!!!
logobj2type = {
	LogObjectType.LOGENTRY : LogEntry,
	LogObjectType.PROXYDATA : ProxyData,
	LogObjectType.CONNECTION : Connection,
	LogObjectType.CONNECTION_OPENED : ConnectionOpened,
	LogObjectType.CONNECTION_CLOSED : ConnectionClosed,
	LogObjectType.CREDENTIAL : Credential,
	LogObjectType.POISONRESULT : PoisonResult,
	LogObjectType.EMAILENTRY : EmailEntry,
	LogObjectType.TRAFFICLOG : TrafficLog,
	LogObjectType.REMOTELOG : RemoteLog,
}

logobj2type_inv = {v: k for k, v in logobj2type.items()}

def get_rdns_tld(rdns):
	if not rdns:
		return None
	m = rdns.rfind(".")
	if m == -1:
		return None
	return rdns[m+1:]

class UnifiedLog:
	def __init__(self):
		self.log_type = None #local or remote
		self.client_id = None

		self.connection_id = None
		self.remote_ip   = None
		self.remote_dns  = None
		self.remote_dns_tld  = None
		self.remote_port = None
		self.local_ip    = None
		self.local_port  = None
		self.connection_timestamp = None

		self.logentry = None
		self.proxydata = None
		self.credential = None
		self.poisonresult = None
		self.email = None
		self.connection_opened = None
		self.connection_closed = None
		self.traffic = None

	@staticmethod
	def construct(log_obj):
		ul = UnifiedLog()

		if isinstance(log_obj, RemoteLog):
			ul.log_type = 'REMOTE'
			ul.client_id = log_obj.client_id

			log_obj = log_obj.log_obj

			if isinstance(log_obj, LogEntry):
				ul.logentry = log_obj

			else:
				ul.connection_id = log_obj.connection.connection_id
				ul.remote_ip   = log_obj.connection.remote_ip
				ul.remote_dns  = log_obj.connection.remote_dns
				ul.remote_port = log_obj.connection.remote_port
				ul.local_ip    = log_obj.connection.local_ip
				ul.local_port  = log_obj.connection.local_port
				ul.connection_timestamp = log_obj.connection.timestamp
				ul.remote_dns_tld = get_rdns_tld(ul.remote_dns)


		else:
			#TODO! check for local IP-port data
			ul.log_type = 'LOCAL'
			ul.local_ip    = None
			ul.local_port  = None

			if isinstance(log_obj, LogEntry):
				ul.logentry = log_obj

			else:
				#logentry doesnt have connection info
				ul.connection_id = log_obj.connection.connection_id
				ul.remote_ip   = log_obj.connection.remote_ip
				ul.remote_dns  = log_obj.connection.remote_dns
				ul.remote_port = log_obj.connection.remote_port
				ul.connection_timestamp = log_obj.connection.timestamp
				ul.remote_dns_tld  = get_rdns_tld(ul.remote_dns)


		if isinstance(log_obj, ProxyData):
			ul.proxydata = log_obj

		elif isinstance(log_obj, Credential):
			ul.credential = log_obj

		elif isinstance(log_obj, PoisonResult):
			ul.poisonresult = log_obj

		elif isinstance(log_obj, EmailEntry):
			ul.email = log_obj

		elif isinstance(log_obj, ConnectionOpened):
			ul.connection_opened = log_obj

		elif isinstance(log_obj, ConnectionClosed):
			ul.connection_closed = log_obj

		elif isinstance(log_obj, TrafficLog):
			ul.traffic = log_obj

		return ul

	def to_dict(self):
		t = {}
		t['log_type'] = self.log_type
		t['client_id'] = self.client_id
		t['connection_id'] = self.connection_id
		t['remote_ip'] = self.remote_ip
		t['remote_dns'] = self.remote_dns
		t['remote_dns_tld'] = self.remote_dns_tld
		t['remote_port'] = self.remote_port
		t['local_ip'] = self.local_ip
		t['local_port'] = self.local_port
		t['connection_timestamp'] = self.connection_timestamp
		t['logentry'] = self.logentry.to_dict() if self.logentry is not None else None
		t['proxydata'] = self.proxydata.to_dict() if self.proxydata is not None else None
		t['credential'] = self.credential.to_dict() if self.credential is not None else None
		t['poisonresult'] = self.poisonresult.to_dict() if self.poisonresult is not None else None
		t['email'] = self.email.to_dict() if self.email is not None else None
		t['connection_opened'] = self.connection_opened.to_dict() if self.connection_opened is not None else None
		t['connection_closed'] = self.connection_closed.to_dict() if self.connection_closed is not None else None
		t['traffic'] = self.traffic.to_dict() if self.traffic is not None else None

		return t


	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)